/**
 * @file
 * @author Petr Bravenec
 * @copyright GNU Public License
 */

#include "database.h"
#include "msqlquery.h"
#include <QCryptographicHash>
#include <QDebug>
#include <QDir>
#include <QUuid>
#include <math.h>

#ifndef ADMINPASSWORD
#define ADMINPASSWORD "e5a66b96415488a8df63d229d1a3048b29ef757f"
#endif

#ifndef FOTOMONPASSWORD  
#define FOTOMONPASSWORD "ff5bcc26bc440020b177abff256652991c3dd0ef"
#endif

DATABASE::~DATABASE() {
    m_db.close();
    m_dbc.close();
}

#ifdef ARM
#define COMMITINTERVAL 60000
#define SYNCHRONOUS    "OFF"
#define DEFAULTDEVICE  "/dev/ttyA"
#else
#define COMMITINTERVAL 5000
#define SYNCHRONOUS    "FULL"
#define DEFAULTDEVICE  "/dev/ttyS0"
#endif


DATABASE::DATABASE(QSettings *settings, QObject *parent) : QObject(parent) {
    m_in_transaction = false;
    m_timerCommit = new QTimer(this);
    m_timerCommit->setInterval(settings->value("commitinterval", COMMITINTERVAL).toInt());
    m_timerCommit->setSingleShot(false);
    connect(m_timerCommit, SIGNAL(timeout()),
            this,            SLOT(commit()));
    m_timerCommit->start();

    m_synchronous = settings->value("synchronous", SYNCHRONOUS).toString();
    if (m_synchronous != "OFF" &&
        m_synchronous != "NORMAL" &&
        m_synchronous != "FULL" ) {
        m_synchronous = SYNCHRONOUS;
        }

    QString dbdata = settings->value("dbname","fotobot.sqlite3").toString();
    QString dbcopy = settings->value("dbname-copy","fotobot-copy.sqlite3").toString();

    QFile filedbdata(dbdata);
    QFile filedbcopy(dbcopy);
    if (filedbcopy.exists() && !filedbdata.exists()) {
        filedbcopy.copy(dbdata);
        }

    m_db  = open( settings->value("dbname","fotobot.sqlite3").toString() );
    m_dbc = open( settings->value("dbname-copy","fotobot-copy.sqlite3").toString() );
}


/**
 * @brief Opens or creates new database file
 */
QSqlDatabase DATABASE::open(const QString& name) {
    QMutexLocker locker(&m_mutex);
    qDebug() << QString("DATABASE::open(%1)").arg(name) << "current path" << QDir::currentPath();

    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", QUuid::createUuid().toString().toAscii());
    db.setDatabaseName( name );
    db.open();

    QSqlQuery q(db);
    q.exec(QString("pragma synchronous=%1;").arg(m_synchronous));
    q.exec("select * from version;");
    if (!q.next()) {
        create(db);
        }

    upgrade(db);
    return db;
}


/**
 * @brief Commit transactions
 *
 * Since the database is as a default in "autocommit" state,
 * when every standalone query is commited, then the database
 * can be very slow when more lines are connected. For this reason
 * are all inserts grouped in 5 secs block and the commit is
 * requested every 5 secons by the m_timerCommit timer.
 */
void DATABASE::commit() {
    QMutexLocker locker(&m_mutex);
    QSqlQuery q(m_db);
    if (m_in_transaction) {
        q.exec("commit;");
        }
    q.exec("begin;");
    m_in_transaction = true;
    m_timerCommit->start(); // restart timer
}


/**
 * @brief Creates new database structure in opened file
 */
void DATABASE::create(QSqlDatabase& db) {
    qWarning() << "DATABASE::create()";
    QSqlQuery q(db);
    q.exec("create table version (version integer);");
    q.exec("insert into version (version) values (1);");
    q.exec("create table lines ("
            " line                      integer primary key, "
            " description               text, "
            " device                    text, "
            " speed                     integer, "
            " retries                   integer not null default 5, "
            " timeout                   integer not null default 3000, "
            " sn_format                 text "
            " );");
    q.exec("create table invertors ("
            // Primary key, description 
            " invertor                  integer primary key, "
            " description               text, "
            // Address
            " line                      integer, "
            " address                   integer, "
            // What to retrieve and save
            " now_power                 integer, "
            " now_ac_current            integer, "
            " now_ac_voltage            integer, "
            " now_ac_frequency          integer, "
            " now_dc_current            integer, "
            " now_dc_voltage            integer, "
            " day_energy                integer, "
            " day_power_maximum         integer, "
            " day_ac_voltage_maximum    integer, "
            " day_ac_voltage_minimum    integer, "
            " day_dc_voltage_maximum    integer, "
            " day_operating_hours       integer, "
            " total_energy              integer, "
            " total_power_maximum       integer, "
            " total_ac_voltage_maximum  integer, "
            " total_ac_voltage_minimum  integer, "
            " total_dc_voltage_maximum  integer, "
            " total_operating_hours     integer, "
            " unique (line, address) "
            " );");
    q.exec("create table data ("
            " invertor                  integer "
            "                               not null references invertors(invertor) "
            "                               on update cascade on delete cascade, "
            " date                      text, "
            " error                     text, "
            " now_power                 numeric, "
            " now_ac_current            numeric, "
            " now_ac_voltage            numeric, "
            " now_ac_frequency          numeric, "
            " now_dc_current            numeric, "
            " now_dc_voltage            numeric, "
            " day_energy                numeric, "
            " day_power_maximum         numeric, "
            " day_ac_voltage_maximum    numeric, "
            " day_ac_voltage_minimum    numeric, "
            " day_dc_voltage_maximum    numeric, "
            " day_operating_hours       numeric, "
            " total_energy              numeric, "
            " total_power_maximum       numeric, "
            " total_ac_voltage_maximum  numeric, "
            " total_ac_voltage_minimum  numeric, "
            " total_dc_voltage_maximum  numeric, "
            " total_operating_hours     numeric "
            " );");
    q.exec("create index data_idx_invertor_date on data(invertor,date);");

    q.exec("create table users ("
            " user       text primary key, "
            " password   text not null "
            " );");
    q.exec(QString("insert into users (user, password) values ('admin', '%1');").arg(ADMINPASSWORD));

    q.exec("create table maintenance ("
            " days       integer"
            " );");
    q.exec("insert into maintenance (days) values (5);");

    // this should go away in production state
    // for developing purpose only
    q.exec(QString(" insert into lines (line, device, description, retries, timeout, speed) values "
           " (1, '%1', 'Line 1', 5, 3000, 9600);").arg(DEFAULTDEVICE));
    q.exec(" begin;");
    for (int i=1; i<=73; i++) {
        q.exec(QString("insert into invertors (invertor, line, address, now_power) values (%1, 1, %1, 1);").arg(i));
        }
    q.exec(" commit;");

}


void DATABASE::upgrade(QSqlDatabase& db) {
    QSqlQuery q(db);
    q.exec("select version from version;");
    q.next();
    int version = q.value(0).toInt();
    if (version == 1) {
        q.exec("create table interfaceboxes ("
                " interfacebox  integer primary key, "
                " line          integer not null references lines(line) on update cascade on delete cascade, "
                " description   text "
                " );");
        qDebug() << "DATABASE::upgrade() 1 -> 2";
        version++;
        }
    if (version == 2) {
        q.exec("create table geocoordinates ("
                " longitude numeric not null, "
                " latitude  numeric not null "
                " );");
        qDebug() << "DATABASE::upgrade() 2 -> 3";
        version++;
        }
    if (version == 3) {
        q.exec("alter table users add column can_access_data     integer;");
        q.exec("alter table users add column can_change_settings integer;");
        q.exec("delete from users;");
        q.exec(QString("insert into users (user, password, can_access_data, can_change_settings) values ('admin', '%1', 1, 1); ").arg(ADMINPASSWORD));
        q.exec(QString("insert into users (user, password, can_access_data, can_change_settings) values ('fotomon', '%1', 1, 0); ").arg(FOTOMONPASSWORD));
        qDebug() << "DATABASE::upgrade() 3 -> 4";
        version++;
        }

    if (version == 4) {
        q.exec("alter table lines add column type integer;");
        q.exec("alter table lines add column hostname text;");
        q.exec("alter table lines add column portnumber integer;");
        q.exec(QString("update lines set type = %1;").arg(DBT_LINES::Fronius)); 
        qDebug() << "DATABASE::upgrade() 4 -> 5";
        version++;
        }

    if (version == 5) {
        q.exec("alter table invertors add column temperature_1 integer;");
        q.exec("alter table invertors add column temperature_2 integer;");
        q.exec("alter table invertors add column irradiance integer;");
        q.exec("alter table data add column temperature_1 numeric;");
        q.exec("alter table data add column temperature_2 numeric;");
        q.exec("alter table data add column irradiance numeric;");
        qDebug() << "DATABASE::upgrade() 4 -> 5";
        version++;
        }

    if (version == 6) {
        q.exec("alter table data add column status text;");
        qDebug() << "DATABASE::upgrade() 5 -> 6";
        version++;
        }

    q.exec(QString("update version set version=%1;").arg(version));
}


/**
 * @brief Returs list of users or selected user
 */
QList<DBT_USERS> DATABASE::users(const QString& user) {
    QMutexLocker locker(&m_mutex);
    QList<DBT_USERS> list;
    MSqlQuery q(m_db);
    if (user.isEmpty()) {
        q.exec(QString("select user, password, can_access_data, can_change_settings from users;"));
      } else {
        q.exec(QString("select user, password, can_access_data, can_change_settings from users where user='%1';").arg(user));
        }
    while (q.next()) {
        DBT_USERS p;
        int i=0;
        p.user                  = q.value(i++).toString();
        p.password              = q.value(i++).toString();
        p.can_access_data       = q.value(i++).toBool();
        p.can_change_settings   = q.value(i++).toBool();
        list.append(p);
        }
    return list;
}


QString DATABASE::changePassword(DBT_USERS user) {
    QString rc;
    rc = changePassword(m_db,  user);
    rc = changePassword(m_dbc, user);
    return rc;
}

/**
 * @brief Updates user's info (actually only password is changed)
 * 
 * This method can only change existing record, no insert is possible.
 */
QString DATABASE::changePassword(QSqlDatabase& db, DBT_USERS user) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    if (user.password.isEmpty() || user.password == "") {
        return user.user;
        }
    QString passsha1 = QCryptographicHash::hash(user.password.toAscii(), QCryptographicHash::Sha1).toHex();
    q.exec(QString("update users set password='%1' where user='%2'")
        .arg(passsha1)
        .arg(user.user)
        );
    return user.user;
}


void DATABASE::saveUser(DBT_USERS user) {
    saveUser(m_db,  user);
    saveUser(m_dbc, user);
}


/**
 * @brief Updates user's info
 */
void DATABASE::saveUser(QSqlDatabase& db, DBT_USERS user) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);

    q.exec(QString("select * from users where user='%1'").arg(user.id));
    if (q.next()) {
        q.exec(QString("update users set user='%1', can_access_data=%2, can_change_settings=%3 where user='%4'")
            .arg(user.user)
            .arg(nullb(user.can_access_data))
            .arg(nullb(user.can_change_settings))
            .arg(user.id)
            );
      } else {
        q.exec(QString("insert into users (user, password, can_access_data, can_change_settings) values ('%1', '!!', %2, %3);")
            .arg(user.user)
            .arg(user.can_access_data)
            .arg(user.can_change_settings)
            );
        }

    locker.unlock();
    changePassword(user);
}


void DATABASE::deleteUser(QString userid) {
    deleteUser(m_db,  userid);
    deleteUser(m_dbc, userid);
}


void DATABASE::deleteUser(QSqlDatabase& db, QString userid) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    q.exec(QString("delete from users where user='%1'").arg(userid));
}



/**
 * @brief Returns list of communications lines
 *
 * @param line - number of requested line or 0 for all lines in database
 *
 * @returns Called without parameters or with line==0 returns list of all lines
 *   and called with line>0 returns empty list (requested line was not found in database)
 *   or list with exactly one record (requested line was found).
 */
QList<DBT_LINES> DATABASE::lines(int line) {
    QMutexLocker locker(&m_mutex);
    QList<DBT_LINES> x;
    MSqlQuery q(m_db);
    q.exec(QString("select line, device, description, speed, retries, timeout, sn_format, type, hostname, portnumber from lines "
        " where %1 <= 0 or line = %1; ").arg(line));
    while (q.next()) {
        DBT_LINES p;
        int i=0;
        p.line          = q.value(i++).toInt();
        p.device        = q.value(i++).toString();
        p.description   = q.value(i++).toString();
        p.speed         = q.value(i++).toInt();
        p.retries       = q.value(i++).toInt();
        p.timeout       = q.value(i++).toInt();
        p.sn_format     = q.value(i++).toString();
        int type        = q.value(i++).toInt();
        p.type          = (type == DBT_LINES::Fronius) ? DBT_LINES::Fronius 
                        : (type == DBT_LINES::ModbusTCP) ? DBT_LINES::ModbusTCP 
                        : (type == DBT_LINES::ModbusRTU) ? DBT_LINES::ModbusRTU 
                        : DBT_LINES::Unknown;
        p.hostname      = q.value(i++).toString();
        p.portnumber    = q.value(i++).toInt();
        x.append(p);
        }
    return x;
}

/**
 * @brief Returns list of invertors filtered by line number
 */
QList<DBT_INVERTORS> DATABASE::invertors(int line) {
    QMutexLocker locker(&m_mutex);
    QList<DBT_INVERTORS> list;
    MSqlQuery q(m_db);
    q.exec(QString("select invertor, description, line, address, "
             " now_power, now_ac_current, now_ac_voltage, now_ac_frequency, now_dc_current,"
             " now_dc_voltage, day_energy, day_power_maximum, day_ac_voltage_maximum, "
             " day_ac_voltage_minimum, day_dc_voltage_maximum, day_operating_hours, "
             " total_energy, total_power_maximum, total_ac_voltage_maximum, "
             " total_ac_voltage_minimum, total_dc_voltage_maximum, total_operating_hours, "
             " temperature_1, temperature_2, irradiance "
             " from invertors where 0 = %1 or line = %1 order by invertor; ")
            .arg(line)
            );
    while (q.next()) {
        DBT_INVERTORS x;
        int i=0;
        x.invertor                  = q.value(i++).toInt();
        x.description               = q.value(i++).toString();
        x.line                      = q.value(i++).toInt();
        x.address                   = q.value(i++).toInt();
        x.now_power                 = q.value(i++).toBool();
        x.now_ac_current            = q.value(i++).toBool();
        x.now_ac_voltage            = q.value(i++).toBool();
        x.now_ac_frequency          = q.value(i++).toBool();
        x.now_dc_current            = q.value(i++).toBool();
        x.now_dc_voltage            = q.value(i++).toBool();
        x.day_energy                = q.value(i++).toBool();
        x.day_power_maximum         = q.value(i++).toBool();
        x.day_ac_voltage_maximum    = q.value(i++).toBool();
        x.day_ac_voltage_minimum    = q.value(i++).toBool();
        x.day_dc_voltage_maximum    = q.value(i++).toBool();
        x.day_operating_hours       = q.value(i++).toBool();
        x.total_energy              = q.value(i++).toBool();
        x.total_power_maximum       = q.value(i++).toBool();
        x.total_ac_voltage_maximum  = q.value(i++).toBool();
        x.total_ac_voltage_minimum  = q.value(i++).toBool();
        x.total_dc_voltage_maximum  = q.value(i++).toBool();
        x.total_operating_hours     = q.value(i++).toBool();
        x.temperature_1             = q.value(i++).toBool();
        x.temperature_2             = q.value(i++).toBool();
        x.irradiance                = q.value(i++).toBool();
        list.append(x);
        }
    return list;
}


/**
 * @brief Returns list of invertors filtered by invertors numbers
 *
 * @param invertors - list of invertors to select
 *
 * Using:
 *
 * @code
 * db->invertors( QList<int>() << 1 << 2 << 3 );
 * @endcode
 */
QList<DBT_INVERTORS> DATABASE::invertors(const QList<int>& invertors) {
    QMutexLocker locker(&m_mutex);
    QList<DBT_INVERTORS> list;
    QString cond = "0=1 ";
    for (int i=0; i<invertors.size(); i++) {
        cond = cond + QString(" or invertor=%1").arg(invertors.at(i));
        }
    MSqlQuery q(m_db);
    q.exec(QString("select invertor, description, line, address, "
             " now_power, now_ac_current, now_ac_voltage, now_ac_frequency, now_dc_current,"
             " now_dc_voltage, day_energy, day_power_maximum, day_ac_voltage_maximum, "
             " day_ac_voltage_minimum, day_dc_voltage_maximum, day_operating_hours, "
             " total_energy, total_power_maximum, total_ac_voltage_maximum, "
             " total_ac_voltage_minimum, total_dc_voltage_maximum, total_operating_hours, "
             " temperature_1, temperature_2, irradiance "
             " from invertors where %1 order by invertor; ")
            .arg(cond)
            );
    while (q.next()) {
        DBT_INVERTORS x;
        int i=0;
        x.invertor                  = q.value(i++).toInt();
        x.description               = q.value(i++).toString();
        x.line                      = q.value(i++).toInt();
        x.address                   = q.value(i++).toInt();
        x.now_power                 = q.value(i++).toBool();
        x.now_ac_current            = q.value(i++).toBool();
        x.now_ac_voltage            = q.value(i++).toBool();
        x.now_ac_frequency          = q.value(i++).toBool();
        x.now_dc_current            = q.value(i++).toBool();
        x.now_dc_voltage            = q.value(i++).toBool();
        x.day_energy                = q.value(i++).toBool();
        x.day_power_maximum         = q.value(i++).toBool();
        x.day_ac_voltage_maximum    = q.value(i++).toBool();
        x.day_ac_voltage_minimum    = q.value(i++).toBool();
        x.day_dc_voltage_maximum    = q.value(i++).toBool();
        x.day_operating_hours       = q.value(i++).toBool();
        x.total_energy              = q.value(i++).toBool();
        x.total_power_maximum       = q.value(i++).toBool();
        x.total_ac_voltage_maximum  = q.value(i++).toBool();
        x.total_ac_voltage_minimum  = q.value(i++).toBool();
        x.total_dc_voltage_maximum  = q.value(i++).toBool();
        x.total_operating_hours     = q.value(i++).toBool();
        x.temperature_1             = q.value(i++).toBool();
        x.temperature_2             = q.value(i++).toBool();
        x.irradiance                = q.value(i++).toBool();
        list.append(x);
        }
    return list;
}

/**
 * @brief Returns list of interfaceboxes filtered by line number
 */
QList<DBT_INTERFACEBOXES> DATABASE::interfaceboxes(int line) {
    QMutexLocker locker(&m_mutex);
    QList<DBT_INTERFACEBOXES> list;
    MSqlQuery q(m_db);
    q.exec(QString("select interfacebox, line, description "
             " from interfaceboxes where 0 = %1 or line = %1 order by interfacebox; ")
            .arg(line)
            );
    while (q.next()) {
        DBT_INTERFACEBOXES x;
        int i=0;
        x.interfacebox              = q.value(i++).toInt();
        x.line                      = q.value(i++).toInt();
        x.description               = q.value(i++).toString();
        list.append(x);
        }
    return list;
}

/**
 * @brief Returns list of interfaceboxes filtered by interfaceboxes numbers
 *
 * @param interfaceboxes - list of interfaceboxes to select
 *
 * Using:
 *
 * @code
 * db->interfaceboxes( QList<int>() << 1 << 2 << 3 );
 * @endcode
 */
QList<DBT_INTERFACEBOXES> DATABASE::interfaceboxes(const QList<int>& interfaceboxes) {
    QMutexLocker locker(&m_mutex);
    QList<DBT_INTERFACEBOXES> list;
    QString cond = "0=1 ";
    for (int i=0; i<interfaceboxes.size(); i++) {
        cond = cond + QString(" or interfacebox=%1").arg(interfaceboxes.at(i));
        }
    MSqlQuery q(m_db);
    q.exec(QString("select interfacebox, line, description "
             " from interfaceboxes where %1 order by interfacebox; ")
            .arg(cond)
            );
    while (q.next()) {
        DBT_INTERFACEBOXES x;
        int i=0;
        x.interfacebox              = q.value(i++).toInt();
        x.line                      = q.value(i++).toInt();
        x.description               = q.value(i++).toString();
        list.append(x);
        }
    return list;
}


/**
 * @brief Returns geocoordinates
 */
DBT_GEOCOORDINATES DATABASE::geocoordinates() {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(m_db);
    q.exec(QString("select longitude, latitude "
             " from geocoordinates; ")
            );
    q.next();
    DBT_GEOCOORDINATES g;
    int i=0;
    g.longitude = q.value(i++);
    g.latitude  = q.value(i++);
    return g;
}

/**
 * @brief Returns data for all invertors
 *
 * @param from  - Date & Time from
 * @param to    - Date & Time to
 * @param limit - Limit number of records
 *
 * When no parameter was given then last known state of all invertors is returned.
 * When just one parameter was given then all records from this time till last know state is returned.
 * Parameter limit is used to limit the number of records. There is a problem to compress big ammount of 
 * data (fbhw is low memory device).
 *
 * @returns
 * Returns list of DBT_DATA_VW structures. Values are stored in QVariant and null values in database
 * should correspond to null values in QVariant variables in the returned structure.
 */
QList<DBT_DATA_VW> DATABASE::data(const QDateTime& from, const QDateTime& to, bool limit) {
    QMutexLocker locker(&m_mutex);
    QList<DBT_DATA_VW> list;
    MSqlQuery q(m_db);

    QString dotaz;
    if (!from.isValid()) {
        dotaz = QString("select i.line, i.address, "
                "d.date, d.error, d.now_power, d.now_ac_current, d.now_ac_voltage, "
                "d.now_ac_frequency, d.now_dc_current, d.now_dc_voltage, d.day_energy, "
                "d.day_power_maximum, d.day_ac_voltage_maximum, d.day_ac_voltage_minimum, "
                "d.day_dc_voltage_maximum, d.day_operating_hours, d.total_energy, "
                "d.total_power_maximum, d.total_ac_voltage_maximum, d.total_ac_voltage_minimum, "
                "d.total_dc_voltage_maximum, d.total_operating_hours, "
                "d.temperature_1, d.temperature_2, d.irradiance, d.status, "
                "l.sn_format "
                "from data d, "
                "    (select invertor,max(date) as date from data group by invertor) x, "
                "    invertors i, "
                "    lines l "
                "where d.invertor=x.invertor and d.date=x.date "
                "  and d.invertor=i.invertor "
                "  and l.line=i.line "
                "order by i.line, i.address; ");
      } else {
        QDateTime lto = QDateTime::currentDateTime();
        if (to.isValid() && !limit) {
            lto = to;
            }
        if (to.isValid() && limit) {
            lto = from.addSecs(10800);
            if (lto > QDateTime::currentDateTime()) {
                lto = QDateTime::currentDateTime();
                }
            if (lto > to) {
                lto = to;
                }
            }
        dotaz = QString("select i.line, i.address, "
                "d.date, d.error, d.now_power, d.now_ac_current, d.now_ac_voltage, "
                "d.now_ac_frequency, d.now_dc_current, d.now_dc_voltage, d.day_energy, "
                "d.day_power_maximum, d.day_ac_voltage_maximum, d.day_ac_voltage_minimum, "
                "d.day_dc_voltage_maximum, d.day_operating_hours, d.total_energy, "
                "d.total_power_maximum, d.total_ac_voltage_maximum, d.total_ac_voltage_minimum, "
                "d.total_dc_voltage_maximum, d.total_operating_hours, "
                "d.temperature_1, d.temperature_2, d.irradiance, d.status, "
                "l.sn_format "
                "from data d, "
                "    invertors i, "
                "    lines l "
                "where d.invertor=i.invertor "
                "  and l.line=i.line "
                "  and d.date >= '%1' and d.date <= '%2' "
                "order by i.line, d.date, i.address ; ")
                .arg(from.toString("yyyy-MM-dd hh:mm:ss"))
                .arg( lto.toString("yyyy-MM-dd hh:mm:ss"))
                ;
        }

    qDebug() << dotaz;

    q.exec(dotaz);
    while (q.next()) {
        DBT_DATA_VW x;
        int i=0;
        x.line                      = q.value(i++).toInt();
        x.address                   = q.value(i++).toInt();
        x.date                      = QDateTime::fromString(q.value(i++).toString(), "yyyy-MM-dd hh:mm:ss");
        x.error                     = q.value(i++).toString();
        x.now_power                 = q.value(i++);
        x.now_ac_current            = q.value(i++);
        x.now_ac_voltage            = q.value(i++);
        x.now_ac_frequency          = q.value(i++);
        x.now_dc_current            = q.value(i++);
        x.now_dc_voltage            = q.value(i++);
        x.day_energy                = q.value(i++);
        x.day_power_maximum         = q.value(i++);
        x.day_ac_voltage_maximum    = q.value(i++);
        x.day_ac_voltage_minimum    = q.value(i++);
        x.day_dc_voltage_maximum    = q.value(i++);
        x.day_operating_hours       = q.value(i++);
        x.total_energy              = q.value(i++);
        x.total_power_maximum       = q.value(i++);
        x.total_ac_voltage_maximum  = q.value(i++);
        x.total_ac_voltage_minimum  = q.value(i++);
        x.total_dc_voltage_maximum  = q.value(i++);
        x.total_operating_hours     = q.value(i++);
        x.temperature_1             = q.value(i++);
        x.temperature_2             = q.value(i++);
        x.irradiance                = q.value(i++);
        x.status                    = q.value(i++).toString();
        x.serial_number             = q.value(i++).toString();
        if (x.serial_number.isEmpty() || x.serial_number=="") {
            x.serial_number = QString("%1").arg(x.address);
          } else {
            x.serial_number = x.serial_number
                                .replace("$L",QString("%1").arg(x.line))
                                .replace("$A",QString("%1").arg(x.address))
                                .replace("$1A",QString("%1").arg(x.address,1,10,QChar('0')))
                                .replace("$2A",QString("%1").arg(x.address,2,10,QChar('0')))
                                .replace("$3A",QString("%1").arg(x.address,3,10,QChar('0')))
                                .replace("$4A",QString("%1").arg(x.address,4,10,QChar('0')))
                                .replace("$5A",QString("%1").arg(x.address,5,10,QChar('0')))
                                .replace("$6A",QString("%1").arg(x.address,6,10,QChar('0')))
                                .replace("$7A",QString("%1").arg(x.address,7,10,QChar('0')))
                                .replace("$8A",QString("%1").arg(x.address,8,10,QChar('0')))
                                .replace("$9A",QString("%1").arg(x.address,9,10,QChar('0')))
                                ;
            }
        list.append(x);
        }

    return list;
}


/**
 * @brief Writes data to database table DATA
 *
 * This slot should be connected to signal INVERTOR::data(DBT_DATA);
 * Data are passed by value let the queueing is possible - this is 
 * because data source and DATABASE object run in different thread.
 */
void DATABASE::insertData(DBT_DATA x) {
    QMutexLocker locker(&m_mutex);
    if (x.invertor <= 0) return;
    MSqlQuery q(m_db);

    q.exec(QString("select 1 from data where invertor=%1 and date=%2;")
            .arg(x.invertor)
            .arg(nulld(x.date))
           );
    if (q.next()) return;

    if (x.error.isEmpty()) {
        q.exec(QString("insert into data ( "
            "invertor, "                        // 1
            "date, "                            // 2
            "error, "                           // 3
            "now_power, "                       // 4
            "now_ac_current, "                  // 5
            "now_ac_voltage, "                  // 6
            "now_ac_frequency, "                // 7
            "now_dc_current, "                  // 8
            "now_dc_voltage, "                  // 9
            "day_energy, "                      // 10
            "day_power_maximum, "               // 11
            "day_ac_voltage_maximum, "          // 12
            "day_ac_voltage_minimum, "          // 13
            "day_dc_voltage_maximum, "          // 14
            "day_operating_hours, "             // 15 
            "total_energy, "                    // 16
            "total_power_maximum, "             // 17
            "total_ac_voltage_maximum, "        // 18
            "total_ac_voltage_minimum, "        // 19
            "total_dc_voltage_maximum, "        // 20
            "total_operating_hours, "           // 21
            "temperature_1, "                   // 22
            "temperature_2, "                   // 23
            "irradiance,"                       // 24
            "status"                            // 25
            ") values ("
            "%1,"                               // invertor
            "%2,"                               // date
            "%3,"                               // error
            "%4,"                               // now_power
            "%5,"                               // now_ac_current
            "%6,"                               // now_ac_voltage
            "%7,"                               // now_ac_frequency
            "%8,"                               // now_dc_current
            "%9,"                               // nod_dc_voltage
            "%10,"                              // day_energy
            "%11,"                              // day_power_maximum
            "%12,"                              // day_ac_voltage_maximum
            "%13,"                              // day_ac_voltage_minimum
            "%14,"                              // day_dc_voltage_maximum
            "%15,"                              // day_operating_hours
            "%16,"                              // total_energy
            "%17,"                              // total_power_maximum
            "%18,"                              // total_ac_voltage_maximum
            "%19,"                              // total_ac_voltage_minimum
            "%20,"                              // total_dc_voltage_maximum
            "%21,"                              // total_operating_hours
            "%22,"                              // temperature_1
            "%23,"                              // temperature_2
            "%24,"                              // irradiance
            "%25"                               // status
            ")" )
            .arg(      x.invertor)                   // 1 
            .arg(nulld(x.date))                      // 2
            .arg(     "null")                        // 3
            .arg(nulli(x.now_power))                 // 4
            .arg(nullf(x.now_ac_current))            // 5
            .arg(nulli(x.now_ac_voltage))            // 6
            .arg(nullf(x.now_ac_frequency))          // 7
            .arg(nullf(x.now_dc_current))            // 8
            .arg(nulli(x.now_dc_voltage))            // 9
            .arg(nulli(x.day_energy))                // 10
            .arg(nulli(x.day_power_maximum))         // 11
            .arg(nulli(x.day_ac_voltage_maximum))    // 12
            .arg(nulli(x.day_ac_voltage_minimum))    // 13
            .arg(nulli(x.day_dc_voltage_maximum))    // 14
            .arg(nullf(x.day_operating_hours))       // 15
            .arg(nulli(x.total_energy))              // 16
            .arg(nulli(x.total_power_maximum))       // 17
            .arg(nulli(x.total_ac_voltage_maximum))  // 18
            .arg(nulli(x.total_ac_voltage_minimum))  // 19
            .arg(nulli(x.total_dc_voltage_maximum))  // 20
            .arg(nullf(x.total_operating_hours))     // 21
            .arg(nullf(x.temperature_1))             // 22
            .arg(nullf(x.temperature_2))             // 23
            .arg(nullf(x.irradiance))                // 24
            .arg(nulls(x.status))                    // 25
            );
    } else {
        // Error, write only error messsage
        q.exec(QString("insert into data ("
            "invertor,"
            "date,"
            "error"
            ") values ("
            "%1,"
            "%2,"
            "%3"
            ")" )
            .arg(      x.invertor)
            .arg(nulld(x.date))
            .arg(nulls(x.error))
            );
        }

}

int DATABASE::saveLine(DBT_LINES x) {
    int rc;
    rc = saveLine(m_db,  x);
    rc = saveLine(m_dbc, x);
    emit databaseChanged();
    return rc;
}

/**
 * @brief Insert or update LINE
 *
 * @returns Number of line in database (primary key)
 */
int DATABASE::saveLine(QSqlDatabase& db, DBT_LINES x) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    q.exec(QString("select line from lines where line=%1;").arg(x.line));
    if (q.next()) {
        // record exists
        q.exec(QString("update lines set "
                        " device='%1', "
                        " description='%2', "
                        " speed=%3, "
                        " retries=%4, "
                        " timeout=%5, "
                        " sn_format='%6', "
                        " type = %7, "
                        " hostname = '%8', "
                        " portnumber = %9 "
                        " where line=%10;")
                .arg(x.device)
                .arg(x.description)
                .arg(x.speed)
                .arg(x.retries)
                .arg(x.timeout)
                .arg(x.sn_format)
                .arg(x.type)
                .arg(x.hostname)
                .arg(x.portnumber)
                .arg(x.line)
                );
      } else {
        q.exec(QString("insert into lines "
                        "(device, description, speed, retries, timeout, sn_format, type, hostname, portnumber) "
                        "values "
                        "('%1', '%2', %3, %4, %5, '%6', %7, '%8', %9);")
                .arg(x.device)
                .arg(x.description)
                .arg(x.speed)
                .arg(x.retries)
                .arg(x.timeout)
                .arg(x.sn_format)
                .arg(x.type)
                .arg(x.hostname)
                .arg(x.portnumber)
                );
        q.exec(QString("select last_insert_rowid();"));
        q.next();
        emit databaseChanged();
        return q.value(0).toInt();
        }
    return x.line;
}


int DATABASE::saveInvertor(DBT_INVERTORS x) {
    int rc;
    rc = saveInvertor(m_db,  x);
    rc = saveInvertor(m_dbc, x);
    emit databaseChanged();
    return rc;
}

/**
 * @brief Insert or update INVERTOR
 *
 * @returns Number of invertor in database (primary key)
 */
int DATABASE::saveInvertor(QSqlDatabase& db, DBT_INVERTORS x) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    q.exec(QString("select line from invertors where invertor=%1;").arg(x.invertor));
    if (q.next()) {
        // record exists
        q.exec(QString("update invertors set "
                "description               = %1, "
                "line                      = %2, "
                "address                   = %3, "
                "now_power                 = %4, "
                "now_ac_current            = %5, "
                "now_ac_voltage            = %6, "
                "now_ac_frequency          = %7, "
                "now_dc_current            = %8, "
                "now_dc_voltage            = %9, "
                "day_energy                = %10, "
                "day_power_maximum         = %11, "
                "day_ac_voltage_maximum    = %12, "
                "day_ac_voltage_minimum    = %13, "
                "day_dc_voltage_maximum    = %14, "
                "day_operating_hours       = %15, "
                "total_energy              = %16, "
                "total_power_maximum       = %17, "
                "total_ac_voltage_maximum  = %18, "
                "total_ac_voltage_minimum  = %19, "
                "total_dc_voltage_maximum  = %20, "
                "total_operating_hours     = %21, "
                "temperature_1             = %22, "
                "temperature_2             = %23, "
                "irradiance                = %24 "
                "where invertor = %25; ")
                .arg(nulls(x.description))
                .arg(nulli(x.line))
                .arg(nulli(x.address))
                .arg(nullb(x.now_power))
                .arg(nullb(x.now_ac_current))
                .arg(nullb(x.now_ac_voltage))
                .arg(nullb(x.now_ac_frequency))
                .arg(nullb(x.now_dc_current))
                .arg(nullb(x.now_dc_voltage))
                .arg(nullb(x.day_energy))
                .arg(nullb(x.day_power_maximum))
                .arg(nullb(x.day_ac_voltage_maximum))
                .arg(nullb(x.day_ac_voltage_minimum))
                .arg(nullb(x.day_dc_voltage_maximum))
                .arg(nullb(x.day_operating_hours))
                .arg(nullb(x.total_energy))
                .arg(nullb(x.total_power_maximum))
                .arg(nullb(x.total_ac_voltage_maximum))
                .arg(nullb(x.total_ac_voltage_minimum))
                .arg(nullb(x.total_dc_voltage_maximum))
                .arg(nullb(x.total_operating_hours))
                .arg(nulli(x.temperature_1))
                .arg(nulli(x.temperature_2))
                .arg(nulli(x.irradiance))
                .arg(nulli(x.invertor))

                );
      } else {
        q.exec(QString("insert into invertors ("
                "description, line, address, now_power, now_ac_current, now_ac_voltage, "
                "now_ac_frequency, now_dc_current, now_dc_voltage, day_energy, "
                "day_power_maximum, day_ac_voltage_maximum, day_ac_voltage_minimum, "
                "day_dc_voltage_maximum, day_operating_hours, total_energy, "
                "total_power_maximum, total_ac_voltage_maximum, total_ac_voltage_minimum, "
                "total_dc_voltage_maximum, total_operating_hours, temperature_1, temperature_2, irradiance) values ("
                "%1, %2, %3, %4, %5, %6, %7, %8, %9, %10, "
                "%11, %12, %13, %14, %15, %16, %17, %18, %19, %20, %21, %22, %23, %24);")
                .arg(nulls(x.description))
                .arg(nulli(x.line))
                .arg(nulli(x.address))
                .arg(nullb(x.now_power))
                .arg(nullb(x.now_ac_current))
                .arg(nullb(x.now_ac_voltage))
                .arg(nullb(x.now_ac_frequency))
                .arg(nullb(x.now_dc_current))
                .arg(nullb(x.now_dc_voltage))
                .arg(nullb(x.day_energy))
                .arg(nullb(x.day_power_maximum))
                .arg(nullb(x.day_ac_voltage_maximum))
                .arg(nullb(x.day_ac_voltage_minimum))
                .arg(nullb(x.day_dc_voltage_maximum))
                .arg(nullb(x.day_operating_hours))
                .arg(nullb(x.total_energy))
                .arg(nullb(x.total_power_maximum))
                .arg(nullb(x.total_ac_voltage_maximum))
                .arg(nullb(x.total_ac_voltage_minimum))
                .arg(nullb(x.total_dc_voltage_maximum))
                .arg(nullb(x.total_operating_hours))
                .arg(nulli(x.temperature_1))
                .arg(nulli(x.temperature_2))
                .arg(nulli(x.irradiance))
                );
        q.exec(QString("select last_insert_rowid();"));
        q.next();
        emit databaseChanged();
        return q.value(0).toInt();
        }
    return x.invertor;
}


int DATABASE::saveInterfacebox(DBT_INTERFACEBOXES x) {
    int rc;
    rc = saveInterfacebox(m_db,  x);
    rc = saveInterfacebox(m_dbc, x);
    emit databaseChanged();
    return rc;
}

/**
 * @brief Insert or update Interfacebox
 *
 * @returns Number of interfaceboxes in database (primary key)
 */
int DATABASE::saveInterfacebox(QSqlDatabase& db, DBT_INTERFACEBOXES x) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    q.exec(QString("select line from interfaceboxes where interfacebox=%1;").arg(x.interfacebox));
    if (q.next()) {
        // record exists
        q.exec(QString("update interfaceboxes set "
                "description        = %1, "
                "line               = %2 "
                "where interfacebox = %3; ")
                .arg(nulls(x.description))
                .arg(nulli(x.line))
                .arg(nulli(x.interfacebox))
                );
      } else {
        q.exec(QString("insert into interfaceboxes ("
                "description, line) values ("
                "%1, %2);")
                .arg(nulls(x.description))
                .arg(nulli(x.line))
                );
        q.exec(QString("select last_insert_rowid();"));
        q.next();
        emit databaseChanged();
        return q.value(0).toInt();
        }
    return x.interfacebox;
}


void DATABASE::setGeocoordinates(DBT_GEOCOORDINATES x) {
    setGeocoordinates(m_db,  x);
    setGeocoordinates(m_dbc, x);
    emit databaseChanged();
}

/**
 * @brief update geocoordinates
 */
void DATABASE::setGeocoordinates(QSqlDatabase& db, DBT_GEOCOORDINATES x) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    q.exec(QString("select longitude, latitude from geocoordinates;"));
    if (q.next()) {
        // record exists
        q.exec(QString("update geocoordinates set "
                "longitude  = %1, "
                "latitude   = %2 ")
                .arg(nullf(x.longitude))
                .arg(nullf(x.latitude))
                );
    } else {
        q.exec(QString("insert into geocoordinates  ("
                "longitude, latitude) values ("
                "%1, %2);")
                .arg(nullf(x.longitude))
                .arg(nullf(x.latitude))
                );
    }
}


void DATABASE::deleteLine(int line) {
    deleteLine(m_db,  line);
    deleteLine(m_dbc, line);
    emit databaseChanged();
}

/**
 * @brief Delete line in database
 */
void DATABASE::deleteLine(QSqlDatabase& db, int line) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    q.exec(QString("begin;"));
    q.exec(QString("delete from invertors where line=%1;").arg(line));
    q.exec(QString("delete from lines     where line=%1;").arg(line));
    q.exec(QString("commit;"));
}


void DATABASE::deleteInvertor(int invertor) {
    deleteInvertor(m_db,  invertor);
    deleteInvertor(m_dbc, invertor);
    emit databaseChanged();
}


/**
 * @brief Delete invertor in database
 */
void DATABASE::deleteInvertor(QSqlDatabase& db, int invertor) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    q.exec(QString("delete from invertors where invertor=%1;").arg(invertor));
}


void DATABASE::deleteInterfacebox(int interfacebox) {
    deleteInterfacebox(m_db,  interfacebox);
    deleteInterfacebox(m_dbc, interfacebox);
    emit databaseChanged();
}

/**
 * @brief Delete interfacebox in database
 */
void DATABASE::deleteInterfacebox(QSqlDatabase& db, int interfacebox) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    q.exec(QString("delete from interfaceboxes where interfacebox=%1;").arg(interfacebox));
}


/**
 * @brief Database maintenance
 *
 * Deletes old records from "data" database table
 * The age for deleting is taken from database table "maintenance".
 */
void DATABASE::execMaintenance() {
    QMutexLocker locker(&m_mutex);
    qWarning() << "DATABASE::execMaintenance();";
    MSqlQuery q(m_db);
    q.exec(QString("select days from maintenance;"));
    if (!q.next()) return;
    QDateTime date;
    date = QDateTime::currentDateTime();
    date = date.addDays(-q.value(0).toInt());
    q.exec(QString("delete from data where date < '%1';")
            .arg(date.toString("yyyy-MM-dd hh:mm:ss"))
            );
    q.exec(QString("vacuum full;"));
}


/**
 * @brief Returns status of lines
 *
 * Status consists from two basic information:
 * - number of invertors connected to the line
 * - number of working invertors. 
 * 
 * Working invertor is invertor which int last 10 minutes
 * has at least one record in database without error.
 *
 * Status is returned as a list of DBT_LINES_STATUS structures.
 */
QList<DBT_LINES_STATUS> DATABASE::linesStatus() {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(m_db);
    QDateTime date = QDateTime::currentDateTime().addSecs(-600);
    QList<DBT_LINES_STATUS> list;
    QString dotaz = QString("select n.line, w.working, n.number "
            "   from (select line, count(0) as number "
            "           from invertors group by line) n "
            "   left join (select line, coalesce(sum(jedna),0) as working "
            "                   from (select i.line, 1 as jedna "
            "                       from data d, invertors i "
            "                       where d.invertor=i.invertor "
            "                       and date>'%1' "
            "                       and error is null "
            "                       group by i.line, i.invertor "
            "                       having count(1) > 0) "
            "                   group by line) w using(line);")
                .arg(date.toString("yyyy-MM-dd hh:mm:ss")
            );
    q.exec(dotaz);
//  qDebug() << dotaz.replace(QRegExp("\\s+"), QString(" "));
    while  (q.next()) {
        DBT_LINES_STATUS p;
        int i=0;
        p.line      = q.value(i++).toInt();
        p.working   = q.value(i++).toInt();
        p.invertors = q.value(i++).toInt();
        list.append(p);
        }
    return list;
}


/**
 * @brief Returns list of DBT_MAINTENANCE
 */
QList<DBT_MAINTENANCE> DATABASE::maintenance() {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(m_db);
    QList<DBT_MAINTENANCE> list;
    q.exec(QString("select days from maintenance;"));
    while (q.next()) {
        DBT_MAINTENANCE p;
        int i=0;
        p.days  = q.value(i++).toInt();
        list.append(p);
        }
    return list;
}


void DATABASE::saveMaintenance(DBT_MAINTENANCE x) {
    saveMaintenance(m_db,  x);
    saveMaintenance(m_dbc, x);
}


/**
 * @brief Updates maintenace record in database
 */
void DATABASE::saveMaintenance(QSqlDatabase& db, DBT_MAINTENANCE x) {
    QMutexLocker locker(&m_mutex);
    MSqlQuery q(db);
    q.exec(QString("update maintenance set days=%1;")
            .arg(nulli(x.days))
            );
}


/**
 * @brief Formats QDateTime for database store
 *
 * Returns string "null" when date is not valid.
 * Returns datetime formated in: 'yyyy-MM-dd hh:mm:ss'
 * include apostrophs.
 */
QString DATABASE::nulld(const QDateTime& x) {
    if (!x.isValid()) {
        return "null";
        }
    return QString("'%1'").arg(x.toString("yyyy-MM-dd hh:mm:ss"));
}


/**
 * @brief Formats QVariant - integer for database store
 *
 * Returns string "null" when QVariant is not valid else
 * returns integer as a string.
 */
QString DATABASE::nulli(const QVariant& x) {
    if (!x.isValid()) {
        return "null";
        }
    int val = x.toInt();
    if (isnan(val)) {
        return "null";
        }
    return QString("%1").arg(val);
}


/**
 * @brief Formats QVariant - double for database store
 *
 * Returns string "null" when QVariant is not valid else
 * returns double as a string.
 */
QString DATABASE::nullf(const QVariant& x) {
    if (!x.isValid()) {
        return "null";
        }
    double val = x.toInt();
    if (isnan(val)) {
        return "null";
        }
    return QString("%1").arg(val).replace(",",".");
}


/**
 * @brief Formats QString for database store
 *
 * Returns string "null" when string is not valid else
 * returns string include apostrophs.
 */
QString DATABASE::nulls(const QString& x) {
    if (x.isEmpty()) {
        return "null";
        }
    QString xx = x;
    return QString("'%1'").arg(xx.replace('\'',"''"));
}


/**
 * @brief Formats bool for database store
 *
 * Returns string "null" when string is not valid else
 * returns string include apostrophs.
 */
QString DATABASE::nullb(bool x) {
    return QString("%1").arg((x)?1:0);
}

