/**
 * @file
 * @author Petr Bravenec
 * @copyright GNU Public License
 */

#ifndef _DATABASE_H_
#define _DATABASE_H_

#include <QObject>
#include <QSettings>
#include <QSqlDatabase>
#include <QMutex>
#include <QTimer>
#include "database_tables.h"


/**
 * @brief Class for communication with database
 *
 * The only object that can read and write to the database files. Contains all
 * access methods needed for data storing in the rest of application.
 *
 * Database engine SQLite is single-thread only. When using in multi-thread application
 * there must be only single thread accessing object DATABASE at the same time and only one object
 * can exist in application. 
 *
 * All functions in the DATABASE object are locked using QMutex.
 *
 * <b>Reading from database:</b><br>
 * You can freely read data from database since the mutex locks object and no other
 * thread can access the database at the same time.
 *
 * <b>Writing to database:</b><br>
 * You can freely write data to your database since the mutex locks object and no other
 * thread can acces the database at the same time. All insert/update functions are 
 * implemented as slots so if your thread cannot wait for 
 * database, you can use queued SIGNAL/SLOT mechanism from different threads:
 *
 * @code
 * connect(writer, SIGNAL(data(DBT_DATA)),
 *         db,       SLOT(insertData(DBT_DATA)),
 *         Qt::QueuedConnection);
 * @endcode
 *
 */
class DATABASE : public QObject {
    Q_OBJECT
  public:
   ~DATABASE();
    DATABASE(QSettings *settings, QObject *parent = NULL);

    QList<DBT_LINES>            lines(int line = 0);
    QList<DBT_INVERTORS>        invertors(int line = 0);
    QList<DBT_INVERTORS>        invertors(const QList<int>& invertors);
    QList<DBT_INTERFACEBOXES>   interfaceboxes(int line = 0);
    QList<DBT_INTERFACEBOXES>   interfaceboxes(const QList<int>& interfaceboxes);
    DBT_GEOCOORDINATES          geocoordinates();
    QList<DBT_DATA_VW>          data(const QDateTime& from = QDateTime(), const QDateTime& to = QDateTime(), bool limit = true);
    QList<DBT_USERS>            users(const QString& user = QString());
    QList<DBT_MAINTENANCE>      maintenance();
    QList<DBT_LINES_STATUS>     linesStatus();

  signals:
    /**
     * @brief Signals when information about lines or invertors changed
     *
     * This signal is be connected to INVERTORS object where list of lines and invertors
     * is stored. When the signal is received then all invertors should be re-read from
     * database and recreated.
     */
    void            databaseChanged();

  public slots:
    void            insertData(DBT_DATA);
    void            execMaintenance();
    void            deleteLine(QSqlDatabase&, int);
    void            deleteLine(               int);
    void            deleteInvertor(QSqlDatabase&, int);
    void            deleteInvertor(               int);
    void            deleteInterfacebox(QSqlDatabase&, int);
    void            deleteInterfacebox(               int);
    void            deleteUser(QSqlDatabase&, QString userid);
    void            deleteUser(               QString userid);
    QString         changePassword(QSqlDatabase&, DBT_USERS);
    QString         changePassword(               DBT_USERS);
    void            saveUser(QSqlDatabase&, DBT_USERS);
    void            saveUser(               DBT_USERS);
    int             saveLine(QSqlDatabase&, DBT_LINES);
    int             saveLine(               DBT_LINES);
    int             saveInvertor(QSqlDatabase&, DBT_INVERTORS);
    int             saveInvertor(               DBT_INVERTORS);
    int             saveInterfacebox(QSqlDatabase&, DBT_INTERFACEBOXES);
    int             saveInterfacebox(               DBT_INTERFACEBOXES);
    void            saveMaintenance(QSqlDatabase&, DBT_MAINTENANCE);
    void            saveMaintenance(               DBT_MAINTENANCE);
    void            setGeocoordinates(QSqlDatabase&, DBT_GEOCOORDINATES);
    void            setGeocoordinates(               DBT_GEOCOORDINATES);
    void            commit();

  private:
    bool            m_in_transaction;
    QTimer         *m_timerCommit;
    QSqlDatabase    m_db;
    QSqlDatabase    m_dbc;
    void            create(QSqlDatabase& db);
    void            upgrade(QSqlDatabase& db);
    QSqlDatabase    open(const QString& dbname);
    QString         nulld(const QDateTime& x);
    QString         nulli(const QVariant& x);
    QString         nullf(const QVariant& x);
    QString         nulls(const QString& x);
    QString         nullb(bool x);
    QMutex          m_mutex; ///< Mutex for locking between threads

    QString         m_synchronous;
};

#endif
