/**
 * @file
 * @author Petr Bravenec
 * @copyright GNU Public License
 */

#include "fronius.h"
#include "fronius_request.h"
#include <QCoreApplication>
#include <QDebug>
#include <QEventLoop>
#include "string.h"


FRONIUS::~FRONIUS() {
}

/**
 * @brief Constructor. The very basic initializations.
 *
 * Only basic initializations can be made here. All needed objects must be created
 * in slotInit() function. Constructor runs in different thread than the rest of the class!
 */
FRONIUS::FRONIUS() : INVERTOR() {
    m_port = NULL;
}


/**
 * @brief Function called within running thread to initialize all needed child objects.
 *
 * This is the first function called when the object is created and new thread event loop
 * is started.
 * All needed objects must be created in this function.
 */
void FRONIUS::slotInit() {
    m_timer = new QTimer(this);
    m_timer->setInterval(10);
    m_timer->setSingleShot(true);
    connect(m_timer, SIGNAL(timeout()),
            this,      SLOT(loop()));

    m_number_of_err=0;
    m_number_of_ok=0;
    m_current_invertor_index = 0;

    open();
}


void FRONIUS::slotQuit() {
    if (m_port) {
        m_port->close();
        delete m_port;
        }
}

/**
 * @brief Opens serial port
 *
 * Opens serial port and set all communication parameters.
 * Starts timer for main loop when successfull.
 *
 * Sends the data signal with error message set when an error occured.
 */
void FRONIUS::open() {
    BaudRateType speed = BAUD9600;
    switch (m_line.speed) {
        case    110: speed = BAUD110; break;
        case    300: speed = BAUD300; break;
        case    600: speed = BAUD600; break;
        case   1200: speed = BAUD1200; break;
        case   2400: speed = BAUD2400; break;
        case   4800: speed = BAUD4800; break;
        case   9600: speed = BAUD9600; break;
        case  19200: speed = BAUD19200; break;
        case  38400: speed = BAUD38400; break;
        case  57600: speed = BAUD57600; break;
        case 115200: speed = BAUD115200; break;
        }
    PortSettings spset;
    spset.BaudRate    = speed;
    spset.DataBits    = DATA_8;
    spset.Parity      = PAR_NONE;
    spset.StopBits    = STOP_1;
    spset.FlowControl = FLOW_OFF;
    spset.Timeout_Millisec = m_line.timeout;

    m_port = new QextSerialPort(m_line.device, spset, QextSerialPort::Polling);
    m_port->open(QIODevice::ReadWrite);
    if (m_port == NULL || !m_port->isOpen()) {
        // Send the signal data() for every invertor with error message and return
        DBT_DATA err;
        err.error = tr("Could not open %1: %2").arg(m_line.device).arg(m_port->errorString());
        for (int i=0; i<m_invertors.size(); i++) {
            err.invertor = m_invertors.at(i).invertor;
            emit data(err);
            }
        setStatus(
                0,
                0,
                tr("Open"),
                tr("Could not open serial port %1").arg(m_line.device)
                );
        QTimer::singleShot(3333, this, SLOT(open()));
        return;
        }

    m_port->flush();
    m_timer->start();
    m_retries = m_line.retries;
}


/**
 * @brief Loops the invertors's list
 *
 * Slot is called from timer's signal.
 * Loops throw the invertors's list and call readInvertor() function for every 
 * invertor in the list, one invertor per call.
 *
 * Current invertor index in m_invertors list is stored in m_current_invertor_index member variable.
 *
 * At the end starts timer again so the event loop calls this function again
 * when the inverval passes.
 */
void FRONIUS::loop() {
    if (m_invertors.size() <= 0) return;

    readInvertor();

    m_current_invertor_index++;
    if (m_invertors.size() <= m_current_invertor_index) {
        int secs = 3;
        m_current_invertor_index = 0;
        setStatus(
                0,
                0,
                tr("No command"),
                tr("Sleeping %1 secs to begin new reading cycle").arg(secs)
                );
        m_timer->setInterval(secs*1000); // wait a few seconds after whole cycle
        emit loopFinished(m_number_of_ok, m_number_of_err);
        m_number_of_err=0;
        m_number_of_ok=0;
      } else {
        setStatus(
                m_invertors.at(m_current_invertor_index).address,
                m_retries-m_nr+1,
                decodeCommand(m_command),
                tr("Going to read next invertor")
                );
        m_timer->setInterval(10);
        }

    m_timer->start();
}


/**
 * @brief Reads and checks response read from invertor
 *
 * When an error occured then throws an exception with error description.
 * Data read are stored in response object.
 */
void FRONIUS::readResponse(FRONIUS_response& response) {
    int n;

    // Read header of the datagram
    setStatus(
            m_invertors.at(m_current_invertor_index).address,
            m_retries-m_nr+1,
            decodeCommand(m_command),
            tr("Reading header from serial port")
            );
    n = m_port->read((char *)&response, FRONIUS_response::HEADER_LENGTH );

    // Check header length
    if (n != FRONIUS_response::HEADER_LENGTH) {
        throw tr("Header invalid: %1 bytes read, should be %2.")
            .arg(n)
            .arg(FRONIUS_response::HEADER_LENGTH)
            ;
        }

    // Check header structure
    if (!response.isHeaderValid()) {
        throw tr("Header invalid.");
        }

    // Read rest of the datagram
    setStatus(
            m_invertors.at(m_current_invertor_index).address,
            m_retries-m_nr+1,
            decodeCommand(m_command),
            tr("Reading rest of datagram from serial port")
            );
    n = m_port->read((char *)response.buffer.buffer, response.length+1);

    // Is the checksum ok?
    if (!response.isChecksumValid()) {
        throw tr("Response checksum error");
        }

    // Did the right invertor answered?
    int address = m_invertors.at(m_current_invertor_index).address;
    if (address >= 1000) {
        address = address - 1000;
        }
    if (response.device_number != address) {
        throw tr("Wrong invertor address in response: should be: %1, read: %2")
            .arg(m_invertors.at(m_current_invertor_index).address)
            .arg(response.device_number)
            ;
        }

    // Is there an error?
    if (response.command == FRONIUS_ERROR) {
        QString errmsg = decodeError((FRONIUS::Error  )response.buffer.Error.errorInformation);
        throw tr("Error: %1").arg(errmsg);
        }
}


/**
 * @brief Reads one value from invertor
 *
 * @params command - requested parameter to read from fronius invertor
 *
 * The function complete the request datagram to set to invertor and write
 * it to serial port. All data are the flushed to serial line.
 * Then the function reads and checks the response from invertor.
 * When successfull, the read value is decoded and returned in QVariant.
 *
 * The exception is thrown when an error occured.
 */
QVariant FRONIUS::readValue(FRONIUS::Command command) {
    // Store current command
    m_command = command;

    // Create FRONIUS_request datagram and write it to the serial port and flush all data
    setStatus(
            m_invertors.at(m_current_invertor_index).address,
            m_retries-m_nr+1,
            decodeCommand(m_command),
            tr("Writing command to serial port")
            );
    int address = m_invertors.at(m_current_invertor_index).address;
    if (address >= 1000) {
        address = address - 1000;
        }

//  m_port->read(32); // Přečíst případné nesmysly

    FRONIUS_request request( address, command );
    m_port->write((char *)&request, sizeof(request));
    m_port->flush();

    QEventLoop eventLoop;
    QTimer::singleShot(300, &eventLoop, SLOT(quit()));
    eventLoop.exec();

    // Read response
    FRONIUS_response response;
    readResponse (response);

    // Decode resonse and decode values
    return decodeResponse(response, command);
}


/**
 * @brief Decode response from invertor and returnsi retrieved value
 *
 */
QVariant FRONIUS::decodeResponse(FRONIUS_response& response, FRONIUS::Command command) {
    if (command != response.command) {
        QString reqcmd = decodeCommand((FRONIUS::Command)m_command);
        QString reccmd = decodeCommand(command);
        throw tr("Bad response: requested command: %1 response: %2 ").arg(reqcmd).arg(reccmd);
        }

    QVariant x;
    try {
        switch (command) {
            case FRONIUS_CMD_GET_VERSION:
                x = QString("%1 %2.%3.%4")
                        .arg(response.buffer.Version.ifcType)
                        .arg(response.buffer.Version.versionMajor)
                        .arg(response.buffer.Version.versionMinor)
                        .arg(response.buffer.Version.versionRelease)
                    ;
                break;
            case FRONIUS_CMD_GET_DEVICE_TYPE:
                x = response.buffer.DeviceType.typeIdentification;
                break;
            case FRONIUS_CMD_GET_DATE_TIME:
                x = response.value().toDouble() / 60.0;
                break;
            case FRONIUS_CMD_GET_ACTIVE_INVERTOR_NUMBER:
            case FRONIUS_CMD_GET_POWER_NOW:
            case FRONIUS_CMD_GET_AC_CURRENT_NOW:
            case FRONIUS_CMD_GET_AC_VOLTAGE_NOW:
            case FRONIUS_CMD_GET_AC_FREQUENCY_NOW:
            case FRONIUS_CMD_GET_DC_CURRENT_NOW:
            case FRONIUS_CMD_GET_DC_VOLTAGE_NOW:
            case FRONIUS_CMD_GET_ENERGY_DAY:
            case FRONIUS_CMD_GET_MAXIMUM_POWER_DAY:
            case FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_DAY:
            case FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_DAY:
            case FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_DAY:
            case FRONIUS_CMD_GET_ENERGY_TOTAL:
            case FRONIUS_CMD_GET_MAXIMUM_POWER_TOTAL:
            case FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_TOTAL:
            case FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_TOTAL:
            case FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_TOTAL:
            case FRONIUS_CMD_GET_TEMPERATURE_1:
            case FRONIUS_CMD_GET_TEMPERATURE_2:
            case FRONIUS_CMD_GET_IRRADIANCE:
                x = response.value();
                break;
            case FRONIUS_CMD_GET_OPERATING_HOURS_DAY:
            case FRONIUS_CMD_GET_OPERATING_HOURS_TOTAL:
                x = response.value().toDouble() / 60.0;
                break;
            case FRONIUS_CMD_GET_ENERGY_YEAR:
            case FRONIUS_CMD_GET_YIELD_YEAR:
            case FRONIUS_CMD_GET_YIELD_DAY:
                break;
            case FRONIUS_CMD_GET_INVERTOR_STATUS:
                // x = decodeStatus((FRONIUS::Status)response.buffer.Status.status);
                // x = response.buffer.Status.status;
                x = decodeStatus((FRONIUS::Status)response.value().toInt());
                break;
            };
        return x;
        }
    catch (QString e) {
        throw tr("Value cannot be recognized: %1")
            .arg(e)
            ;
        }
}


/**
 * @brief Reads values from current invertor
 *
 * Retrieved data are stored in DBT_DATA class and
 * sent to other object using signal data(DBT_DATA);
 *
 * When an error occured then the DBT_DATA::error variable is set and all other data stored
 * id DBT_DATA should be ignored in receiver object.
 *
 * Current invertor's parameters are read from m_invetors list.
 * Reads only data required by parameters set in DBT_INVERTORS when communicating with invertors
 * so you can read very quickly only needed values (the actual power for example).
 */
void FRONIUS::readInvertor() {
    const DBT_INVERTORS& invertor = m_invertors.at(m_current_invertor_index);

    DBT_DATA x;
    x.invertor  = invertor.invertor;

    for (m_nr=m_retries; m_nr>0; m_nr--) {
        try {
            if (invertor.now_power)                 x.now_power                 = readValue(FRONIUS_CMD_GET_POWER_NOW).toDouble();
            if (invertor.now_ac_current)            x.now_ac_current            = readValue(FRONIUS_CMD_GET_AC_CURRENT_NOW).toDouble();
            if (invertor.now_ac_voltage)            x.now_ac_voltage            = readValue(FRONIUS_CMD_GET_AC_VOLTAGE_NOW).toDouble();
            if (invertor.now_ac_frequency)          x.now_ac_frequency          = readValue(FRONIUS_CMD_GET_AC_FREQUENCY_NOW).toDouble();
            if (invertor.now_dc_current)            x.now_dc_current            = readValue(FRONIUS_CMD_GET_DC_CURRENT_NOW).toDouble();
            if (invertor.now_dc_voltage)            x.now_dc_voltage            = readValue(FRONIUS_CMD_GET_DC_VOLTAGE_NOW).toDouble();
            if (invertor.day_energy)                x.day_energy                = readValue(FRONIUS_CMD_GET_ENERGY_DAY).toDouble();
            if (invertor.day_power_maximum)         x.day_power_maximum         = readValue(FRONIUS_CMD_GET_MAXIMUM_POWER_DAY).toDouble();
            if (invertor.day_ac_voltage_maximum)    x.day_ac_voltage_maximum    = readValue(FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_DAY).toDouble();
            if (invertor.day_ac_voltage_minimum)    x.day_ac_voltage_minimum    = readValue(FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_DAY).toDouble();
            if (invertor.day_dc_voltage_maximum)    x.day_dc_voltage_maximum    = readValue(FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_DAY).toDouble();
            if (invertor.day_operating_hours)       x.day_operating_hours       = readValue(FRONIUS_CMD_GET_OPERATING_HOURS_DAY).toDouble();
            if (invertor.total_energy)              x.total_energy              = readValue(FRONIUS_CMD_GET_ENERGY_TOTAL).toDouble();
            if (invertor.total_power_maximum)       x.total_power_maximum       = readValue(FRONIUS_CMD_GET_MAXIMUM_POWER_TOTAL).toDouble();
            if (invertor.total_ac_voltage_maximum)  x.total_ac_voltage_maximum  = readValue(FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_TOTAL).toDouble();
            if (invertor.total_ac_voltage_minimum)  x.total_ac_voltage_minimum  = readValue(FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_TOTAL).toDouble();
            if (invertor.total_dc_voltage_maximum)  x.total_dc_voltage_maximum  = readValue(FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_TOTAL).toDouble();
            if (invertor.total_operating_hours)     x.total_operating_hours     = readValue(FRONIUS_CMD_GET_OPERATING_HOURS_TOTAL).toDouble();
            if (invertor.temperature_1)             x.temperature_1             = readValue(FRONIUS_CMD_GET_TEMPERATURE_1).toDouble();
            if (invertor.temperature_2)             x.temperature_2             = readValue(FRONIUS_CMD_GET_TEMPERATURE_2).toDouble();
            if (invertor.irradiance)                x.irradiance                = readValue(FRONIUS_CMD_GET_IRRADIANCE).toDouble();
            try {
                                                    x.status                    = readValue(FRONIUS_CMD_GET_INVERTOR_STATUS).toString();
                } catch (QString e) {
                    x.status = "UNKNOWN";
                    }
            checkData(x);
            emit data(x);
            m_number_of_ok++;
            break;
            }
        catch (QString e) {
            qDebug() << "Repeated invertor reading:" << x.invertor << "try: " << m_nr;
            if (m_nr>1) {
                setStatus(
                        m_invertors.at(m_current_invertor_index).address,
                        m_retries-m_nr+1,
                        decodeCommand(m_command),
                        tr("Error: %1").arg(e)
                        );
                msleep(1000);
                continue;
                }
            qDebug() << "Repeated invertor reading:" << x.invertor << "try: " << m_nr << "gave up";
            QString cmdmsg = decodeCommand((FRONIUS::Command)m_command);
            x.error = QString("%1 Command: %2").arg(e).arg(cmdmsg);
            emit data(x);
            m_number_of_err++;
            break;
            }
        }
}

/**
 * @brief Basic data check
 */
void FRONIUS::checkData(const DBT_DATA& x) {
    if (x.now_power.toDouble() < 0)     throw tr("Now power < 0");
    if (x.now_power.toDouble() > 15000) throw tr("Now power > 15000");
}


/**
 * @brief Converts fronius command code to text description 
 *
 * @returns QString containing the command description
 */
QString FRONIUS::decodeCommand(FRONIUS::Command command) {
    QString commandString;
    switch (command) {
        case FRONIUS_CMD_GET_VERSION:
            commandString = tr("getVersion");
            break;
        case FRONIUS_CMD_GET_DEVICE_TYPE:
            commandString = tr("getDeviceType");
            break;
        case FRONIUS_CMD_GET_DATE_TIME:
            commandString = tr("getDateTime");
            break;
        case FRONIUS_CMD_GET_ACTIVE_INVERTOR_NUMBER:
            commandString = tr("getActiveInverterNumber");
            break;
        case FRONIUS_CMD_GET_POWER_NOW:
            commandString = tr("getPowerNow");
            break;
        case FRONIUS_CMD_GET_ENERGY_TOTAL:
            commandString = tr("getEnergyTotal");
            break;
        case FRONIUS_CMD_GET_ENERGY_DAY:
            commandString = tr("getEnergyDay");
            break;
        case FRONIUS_CMD_GET_ENERGY_YEAR:
            commandString = tr("getEnergyYear");
            break;
        case FRONIUS_CMD_GET_AC_CURRENT_NOW:
            commandString = tr("getACCurrentNow");
            break;
        case FRONIUS_CMD_GET_AC_VOLTAGE_NOW:
            commandString = tr("getACVoltageNow");
            break;
        case FRONIUS_CMD_GET_AC_FREQUENCY_NOW:
            commandString = tr("getACFrequencyNow");
            break;
        case FRONIUS_CMD_GET_DC_CURRENT_NOW:
            commandString = tr("getDCCurrentNow");
            break;
        case FRONIUS_CMD_GET_DC_VOLTAGE_NOW:
            commandString = tr("getDCVoltageNow");
            break;
        case FRONIUS_CMD_GET_YIELD_DAY:
            commandString = tr("getYieldDay");
            break;
        case FRONIUS_CMD_GET_MAXIMUM_POWER_DAY:
            commandString = tr("getMaximumPowerDay");
            break;
        case FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_DAY:
            commandString = tr("getMaximumACVoltageDay");
            break;
        case FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_DAY:
            commandString = tr("getMinimumACVoltageDay");
            break;
        case FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_DAY:
            commandString = tr("getMaximumDCVoltageDay");
            break;
        case FRONIUS_CMD_GET_OPERATING_HOURS_DAY:
            commandString = tr("getOperatingHoursDay");
            break;
        case FRONIUS_CMD_GET_YIELD_YEAR:
            commandString = tr("getYieldYear");
            break;
        case FRONIUS_CMD_GET_MAXIMUM_POWER_TOTAL:
            commandString = tr("getMaximumPowerTotal");
            break;
        case FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_TOTAL:
            commandString = tr("getMaximumACVoltageTotal");
            break;
        case FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_TOTAL:
            commandString = tr("getMinimumACVoltageTotal");
            break;
        case FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_TOTAL:
            commandString = tr("getMaximumDCVoltageTotal");
            break;
        case FRONIUS_CMD_GET_OPERATING_HOURS_TOTAL:
            commandString = tr("getOperatingHoursTotal");
            break;
        case FRONIUS_CMD_GET_TEMPERATURE_1:
            commandString = tr("getTemperature1");
            break;
        case FRONIUS_CMD_GET_TEMPERATURE_2:
            commandString = tr("getTemperature2");
            break;
        case FRONIUS_CMD_GET_IRRADIANCE:
            commandString = tr("getIrradiance");
            break;
        default:
            commandString = tr("Unknown command");
            break;
        }
    return commandString;
}


/**
 * @brief Converts fronius error code to text description 
 *
 * @returns QString containing the error description
 */
QString FRONIUS::decodeError(FRONIUS::Error errorInformation) {
    QString errorString;
    switch (errorInformation) {
        case FRONIUS_PROTOCOL_ERROR_UNKNOWN_COMMAND:
            errorString = tr("unknown command");
            break;
        case FRONIUS_PROTOCOL_ERROR_TIMEOUT:
            errorString = tr("timeout");
            break;
        case FRONIUS_PROTOCOL_ERROR_INCORRECT_DATA_STRUCTURE:
            errorString = tr("incorrect data structure");
            break;
        case FRONIUS_PROTOCOL_ERROR_COMMAND_QUEUE_FULL:
            errorString = tr("command queue full");
            break;
        case FRONIUS_PROTOCOL_ERROR_DEVICE_OR_OPTION_NA:
            errorString = tr("device or option not available");
            break;
        case FRONIUS_PROTOCOL_ERROR_NO_RESPONSE:
            errorString = tr("no response");
            break;
        case FRONIUS_PROTOCOL_ERROR_SENSOR_ERROR:
            errorString = tr("sensor error");
            break;
        case FRONIUS_PROTOCOL_ERROR_SENSOR_NOT_ACTIVE:
            errorString = tr("sensor not active");
            break;
        case FRONIUS_PROTOCOL_ERROR_INCORRECT_COMMAND:
            errorString = tr("incorrect command");
            break;
        case FRONIUS_PROTOCOL_ERROR_DEVICE_ID_COLLISION:
            errorString = tr("device id collision");
            break;
        default:
            errorString = tr("unknown error information");
            break;
        }
    return errorString;
}


/**
 * @brief Converts fronius error code to text description 
 *
 * @returns QString containing the error description
 */
QString FRONIUS::decodeStatus(FRONIUS::Status statusInformation) {
    QString statusString;
    switch (statusInformation) {
        case FRONIUS_STATUS_STARTUP:
            statusString = "STARTING";
            break;
        case FRONIUS_STATUS_OPERATION:
            statusString = "MPPT";
            break;
        case FRONIUS_STATUS_MANUAL_STANDBY:
            statusString = "STANDBY";
            break;
        case FRONIUS_STATUS_FAILURE:
            statusString = "FAULT";
            break;
        default:
            statusString = "UNKNOWN";
            break;
        }
    return statusString;
}


