/**
 * @file
 * @author Petr Bravenec
 * @copyright GNU Public License
 */

#ifndef _FRONIUS_H_
#define _FRONIUS_H_
#include <QList>
#include <QDateTime>
#include <QTimer>
#include <QVariant>

#include "invertor.h"
#include "fronius_response.h"
#include "qextserialport.h"

class DBT_DATA;

/**
 * @brief Communicates with invertors using serial port
 *
 * The class opens one serial port. If you have more ports, you have to have
 * more instances of the FRONIUS class;
 */
class FRONIUS : public INVERTOR {
    Q_OBJECT
  public:
    FRONIUS();
   ~FRONIUS();

  public slots:
    void            open();

  public:
    /**
     * @brief List of all fronius commands
     */
    enum Command {
        FRONIUS_CMD_GET_VERSION                     = 0x01,
        FRONIUS_CMD_GET_DEVICE_TYPE                 = 0x02,
        FRONIUS_CMD_GET_DATE_TIME                   = 0x03,
        FRONIUS_CMD_GET_ACTIVE_INVERTOR_NUMBER      = 0x04,
        FRONIUS_CMD_GET_POWER_NOW                   = 0x10,
        FRONIUS_CMD_GET_ENERGY_TOTAL                = 0x11,
        FRONIUS_CMD_GET_ENERGY_DAY                  = 0x12,
        FRONIUS_CMD_GET_ENERGY_YEAR                 = 0x13,
        FRONIUS_CMD_GET_AC_CURRENT_NOW              = 0x14,
        FRONIUS_CMD_GET_AC_VOLTAGE_NOW              = 0x15,
        FRONIUS_CMD_GET_AC_FREQUENCY_NOW            = 0x16,
        FRONIUS_CMD_GET_DC_CURRENT_NOW              = 0x17,
        FRONIUS_CMD_GET_DC_VOLTAGE_NOW              = 0x18,
        FRONIUS_CMD_GET_YIELD_DAY                   = 0x19,
        FRONIUS_CMD_GET_MAXIMUM_POWER_DAY           = 0x1a,
        FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_DAY      = 0x1b,
        FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_DAY      = 0x1c,
        FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_DAY      = 0x1d,
        FRONIUS_CMD_GET_OPERATING_HOURS_DAY         = 0x1e,
        FRONIUS_CMD_GET_YIELD_YEAR                  = 0x20,
        FRONIUS_CMD_GET_MAXIMUM_POWER_TOTAL         = 0x26,
        FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_TOTAL    = 0x27,
        FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_TOTAL    = 0x28,
        FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_TOTAL    = 0x29,
        FRONIUS_CMD_GET_OPERATING_HOURS_TOTAL       = 0x2a,
        FRONIUS_CMD_GET_INVERTOR_STATUS             = 0x37,
        FRONIUS_CMD_GET_TEMPERATURE_1               = 0xE0,
        FRONIUS_CMD_GET_TEMPERATURE_2               = 0xE1,
        FRONIUS_CMD_GET_IRRADIANCE                  = 0xE2
        };

  private:

    /**
     * @brief List of all fronius error codes
     */
    enum Error {
        FRONIUS_PROTOCOL_ERROR_UNKNOWN_COMMAND      = 0x01,
        FRONIUS_PROTOCOL_ERROR_TIMEOUT              = 0x02,
        FRONIUS_PROTOCOL_ERROR_INCORRECT_DATA_STRUCTURE =0x03,
        FRONIUS_PROTOCOL_ERROR_COMMAND_QUEUE_FULL   = 0x04,
        FRONIUS_PROTOCOL_ERROR_DEVICE_OR_OPTION_NA  = 0x05,
        FRONIUS_PROTOCOL_ERROR_NO_RESPONSE          = 0x06,
        FRONIUS_PROTOCOL_ERROR_SENSOR_ERROR         = 0x07,
        FRONIUS_PROTOCOL_ERROR_SENSOR_NOT_ACTIVE    = 0x08,
        FRONIUS_PROTOCOL_ERROR_INCORRECT_COMMAND    = 0x09,
        FRONIUS_PROTOCOL_ERROR_DEVICE_ID_COLLISION  = 0x0a
        };

    enum Status {
        FRONIUS_STATUS_STARTUP                      = 0x01,
        FRONIUS_STATUS_OPERATION                    = 0x02,
        FRONIUS_STATUS_MANUAL_STANDBY               = 0x03,
        FRONIUS_STATUS_FAILURE                      = 0x04
        };

  private slots:
    void            slotInit();
    void            slotQuit();
    void            loop();

  private:

    static const int FRONIUS_ERROR = 0x0e;

    void            readInvertor  ();
    QVariant        readValue     (FRONIUS::Command);
    void            readResponse  (FRONIUS_response&);
    QVariant        decodeResponse(FRONIUS_response&, FRONIUS::Command);
    void            checkData     (const DBT_DATA&);

    QString         decodeError   (FRONIUS::Error);
    QString         decodeCommand (FRONIUS::Command);
    QString         decodeStatus  (FRONIUS::Status);

    Command         m_command;                      ///< Current command processed
    int             m_current_invertor_index;       ///< Current index in INVERTOR::m_invertors list
    QextSerialPort *m_port;                         ///< Serial port device
    QTimer         *m_timer;                        ///< Timer for main loop

    int             m_nr;
    int             m_retries;
    int             m_number_of_ok;
    int             m_number_of_err;

};

#endif
