/**
 * @file
 * @author Petr Bravenec
 * @copyright GNU Public License
 */

#ifndef _INVERTOR_H_
#define _INVERTOR_H_

#include <QThread>
#include <QMutex>
#include "database_tables.h"
#include "invertor_status.h"

/**
 * @brief Virtual class for invertor communication
 *
 * Every derived class must implement slotInit() nd slotQuit() methods.
 *
 * As parameters the object gets the line specification in DBT_LINES object and
 * list of invertors connected to this line in DBT_INVERTORS's list.
 * One independent thread is started for every line.
 *
 * When all needed parameters are set the start() function must be called.
 *
 * Data are sent to other object using signal data(DBT_DATA);
 */
class INVERTOR : public QThread {
    Q_OBJECT
  public:
   ~INVERTOR();

    /**
     * @brief Factory makes instances depending on type
     */
    static INVERTOR *create(DBT_LINES::Type);

    INVERTOR_status status(); ///< Returns current status of the line

    /**
      * @brief Sets the parameters for the line. 
      *
      * The DBT_LINES corresponds to database table LINES - all needed
      * parameters should be set in the database using webserver part.
      */
    virtual void    setLine(const DBT_LINES&);

    /**
      * @brief Sets the list of invertors connected to the line.
      *
      * The DBT_INVERTORS corresonds to database table INVERTORS.
      *
      */
    virtual void    setInvertors(const QList<DBT_INVERTORS>&);

    /**
      * @brief Starts the thread
      *
      * Starts the thread's event loop, moves the object to that thread and calls
      * from the event loop ancestor's slotInit() function.
      */
    void            start(Priority priority = InheritPriority);

  signals:
    /**
      * @brief Signal to send retrieved data to other objects.
      *
      * In main thread this signal is connected to DATABASE object and saves the stored
      * data to database file.
      *
      * All retrieved data should be stored in the DBT_DATA class. When an error occured,
      * the error member variable must be set. Then the other values stored in class are ignored.
      *
      */ 
    void            data  (DBT_DATA);

    /**
      * @brief Signal is sent when reading cycle was finished and all invertors were read
      *
      * As a response to this signal the database commit should be requested.
      */
    void            loopFinished(int number_of_ok, int number_of_err);

  public slots:
    /**
      * @brief Quits the running thread
      *
      * From this slot the virtual method slotQuit() is called so the child class
      * can stop its work and close all opened file descriptors.
      */
    void            quit();

  protected slots:
    /**
     * @brief Initializes ancestors.
     *
     * This function is called as a first event from event loop when new thread is started.
     *
     * <b>Important:</b> Do not create any Qt object in derived class's constructor. Make new
     * objects (timers, serial lines and so on)  in the slotInit() implementation. Constructor
     * and slotInit() function are called within different threads so all objects created in
     * constructor makes problem later.
     */
    virtual void    slotInit() = 0;

    /**
     * @brief Quits ancestors.
     *
     * This function is called when the quit signal was received.
     * You should close your serial lines and make other necessary jobs to end the invertor.
     * Do not call QThread::quit() from this function since it is called after your action
     * automatically in INVERTOR::quit() method.
     */
    virtual void    slotQuit() = 0;

  protected:
    INVERTOR();

    /**
     * @brief Stores information about line
     */
    DBT_LINES               m_line;

    /**
     * @brief Stores information about all invertors connected to the line
     */
    QList<DBT_INVERTORS>    m_invertors; 

    /**
     * @brief Set status of line
     */
    void                    setStatus(int address, int retries, const QString& command, const QString& status);

  private:
    /**
     * @brief Line status
     */
    INVERTOR_status         m_status;
    QMutex                  m_status_mutex;

};

#endif
