/**
 * @file
 * @author Petr Bravenec
 * @copyright GNU Public License
 */

#include "modbustcp.h"
#include "socketlist.h"
#include <QCoreApplication>
#include <QTcpSocket>
#include <QDebug>
#include <QStringList>
#include <QEventLoop>
#include <QTimer>
#include "string.h"

#define HEXDEBUG(x) \
    { QStringList str; \
    for (int i=0; i<x.size(); i++) { \
        str << QString("%1").arg( (quint8) x[i], 2, 16, QChar('0')); \
        } \
    qDebug() << str.join(" "); \
    }


MODBUSTCP::~MODBUSTCP() {
}

/**
 * @brief Constructor. The very basic initializations.
 *
 * Only basic initializations can be made here. All needed objects must be created
 * in slotInit() function. Constructor runs in different thread than the rest of the class!
 */
MODBUSTCP::MODBUSTCP() : INVERTOR() {
    // qDebug() << "MODBUSTCP::MODBUSTCP()";
    m_frameid = 0;
    m_number_of_err=0;
    m_number_of_ok=0;
    m_current_invertor_index = 0;
}


/**
 * @brief Function called within running thread to initialize all needed child objects.
 *
 * This is the first function called when the object is created and new thread event loop
 * is started.
 * All needed objects must be created in this function.
 */
void MODBUSTCP::slotInit() {
    qDebug() << "MODBUSTCP::slotInit()" << this;
    m_timer = new QTimer(this);
    m_timer->setInterval(10);
    m_timer->setSingleShot(true);
    connect(m_timer, SIGNAL(timeout()),
            this,      SLOT(loop()));
    m_timer->start();

    try {
        SOCKETLIST->addLine(m_line.line, m_line.hostname, m_line.portnumber);
        SOCKETLIST->open(m_line.line);
        } catch (QString e) {};
}


void MODBUSTCP::slotQuit() {
}


/**
 * @brief Opens serial port
 *
 * Opens serial port and set all communication parameters.
 * Starts timer for main loop when successfull.
 *
 * Sends the data signal with error message set when an error occured.
 */
void MODBUSTCP::open() {
    
}


/**
 * @brief Loops the invertors's list
 *
 * Slot is called from timer's signal.
 * Loops throw the invertors's list and call readInvertor() function for every 
 * invertor in the list, one invertor per call.
 *
 * Current invertor index in m_invertors list is stored in m_current_invertor_index member variable.
 *
 * At the end starts timer again so the event loop calls this function again
 * when the inverval passes.
 */
void MODBUSTCP::loop() {
    if (m_invertors.size() <= 0) return;

    readInvertor();

    m_current_invertor_index++;
    if (m_invertors.size() <= m_current_invertor_index) {
        int secs = 3;
        m_current_invertor_index = 0;
        setStatus(
                m_invertors[m_current_invertor_index].address,
                0,
                QString(),
                tr("Sleeping %1 secs to begin new reading cycle").arg(secs)
                );
        m_timer->setInterval(secs*1000); // wait a few seconds after whole cycle
        emit loopFinished(m_number_of_ok, m_number_of_err);
        m_number_of_err=0;
        m_number_of_ok=0;
      } else {
        setStatus(
                m_invertors.at(m_current_invertor_index).address,
                0,
                QString(),
                tr("Going to read next invertor")
                );
        m_timer->setInterval(10);
        }

    m_timer->start();
}


void MODBUSTCP::readInvertor() {
    try {
        readInvertorInfo(m_invertors[m_current_invertor_index]);
        }
    catch (QString e) {
        DBT_DATA x;
        setStatus(m_invertors[m_current_invertor_index].address, 0, QString(), e);
        x.invertor = m_invertors[m_current_invertor_index].invertor;
        x.error = e;
        emit data(x);
        m_number_of_err++;
        qDebug() << "MODBUSTCP::readInvertor()" << e;
        try {
            SOCKETLIST->open(m_invertors[m_current_invertor_index].line);
            } catch (QString e) {};
        }

}


#define CLEANSOCKET() { \
    socket->flush(); \
    socket->waitForReadyRead(m_line.timeout/10); \
    socket->readAll(); \
    }


QByteArray MODBUSTCP::readModbus(const DBT_INVERTORS& invertor, quint16 regaddress, quint16 regcnt, quint16 expectedLength) {
    incrementFrameId();
    QTcpSocket *socket = SOCKETLIST->socket(invertor.line);
    quint8  unitid = invertor.address;
    quint16 length = 4;
    quint8 *f = (quint8 *)(&m_frameid);
    quint8 *l = (quint8 *)(&length);
    quint16 reg = regaddress - 1;

    QByteArray cmd;
    cmd[0] = *(f+1);
    cmd[1] = *(f+0);
    cmd[2] = 0;
    cmd[3] = 0;
    cmd[4] = *(l+1);
    cmd[5] = *(l+0);
    cmd[6] = unitid;
    cmd[7] = 0x03;
    cmd[8] = (reg >> 8) & 0xff;
    cmd[9] =  reg & 0xff;
    cmd[10] = (regcnt >> 8) & 0xff;
    cmd[11] =  regcnt & 0xff;
    socket->write(cmd);

//  HEXDEBUG(cmd);
    qDebug() << "zapis povelu do TCP address" << invertor.address << "frameid" << m_frameid;

    setStatus(invertor.address, 0, QString(), tr("Writing data to TCP socket"));
    if (!socket->waitForBytesWritten(1000)) {
        throw tr("Could not write data to %1").arg(m_line.hostname); 
        }
    qDebug() << "zapsano";

    QByteArray rec;
    for (int trynumber=0; trynumber <= 5; trynumber++) {
        qDebug() << "cteni z TCP";
        setStatus(invertor.address, 0, QString(), tr("Reading data from TCP socket"));

        if (!socket->waitForReadyRead(m_line.timeout)) {
            CLEANSOCKET();
            throw tr("Could not read data from %1").arg(m_line.hostname); 
            }

        QEventLoop loop;
        QTimer::singleShot(100, &loop, SLOT(quit()));
        loop.exec();

        rec = socket->readAll();
        qDebug() << "precteno" << rec.size();

        uint rtcode = static_cast<uint>(rec[7]); // nedávat do podmínky, nechat v proměnné!
        if (rec.size() == 9 && rtcode == 0x83) {
            CLEANSOCKET();
            int exceptionCode = (int)rec[8];
            QString exceptionText = tr("Modbus protocol exception: 0x83 0x%1 %2")
                        .arg(exceptionCode, 2, 16, QChar('0'))
                        .arg(exceptionCodeToString(exceptionCode))
                        ;
            qDebug() << exceptionText;
            throw exceptionText;
            }

        if (rec.size() < expectedLength) {
            CLEANSOCKET();
            HEXDEBUG(rec);
            throw tr("Bad response: read %1 bytes, expected: %2 bytes").arg(rec.size()).arg(expectedLength);
            }

        int fid   = (int)((unsigned char)rec[0]) * 256 
                  + (int)((unsigned char)rec[1]);
        if (fid != m_frameid && trynumber++ < 5 ) {
            qDebug() << "pozadovana adresa" << m_frameid << "vracena adresa" << fid;
            CLEANSOCKET();
            continue;
            }

        if (trynumber >= 5) {
            CLEANSOCKET();
            throw tr("Bad response: frame id received differs from frame id sent");
            }

        break;
        }

    setStatus(invertor.address, 0, QString(), tr("Decoding data read from TCP socket"));
    int uid   = rec[6];
    if (uid != unitid) {
        throw tr("Bad response: unexpected unit id: %1, expected: %2").arg(uid).arg(unitid);
        }

//  HEXDEBUG(rec);

    return rec;
}


void MODBUSTCP::readInvertorInfo(const DBT_INVERTORS& invertor) {
    qDebug() << "MODBUSTCP::readInvertorInfo() invertor line address: " << invertor.line << invertor.invertor << invertor.address;

    // Zjisti status
    DBT_DATA x;
    QByteArray rec = readModbus(invertor, 
                214,    /* Address of register */
                1,      /* Number of registers to read */
                11 /* Expected length */ );
    int errcode = integer16(rec, 9);
    qDebug() << "Invertor status:" << errcode << stateToString(errcode);

    // Precti data
    int istatus;
    for (int i=0; i<3; i++) {
        rec = readModbus(invertor, 
                40071,  /* Address of first read register */
                60,     /* Number of read register*/
                129     /* Expected length */ );
        istatus = integer16(rec, 103);
        if (istatus != 7) { break; }  // FAULT
        if (errcode != 0) { break; }  // Unknown error
        } 

    // Vyhodnotit status invertoru
    QString status;
    bool working = false;
    switch (istatus) {
        case  1: working = false; status = "OFF"; break;
        case  2: working = false; status = "SLEEPING"; break;
        case  3: working = false; status = "STARTING"; break;
        case  4: working = true;  status = "MPPT"; break;
        case  5: working = true;  status = "THROTTLED"; break;
        case  6: working = false; status = "SHUTTING_DOWN"; break;
        case  7: working = false; status = "FAULT"; break;
        case  8: working = false; status = "STANDBY"; break;
        case  9: working = false; status = "NO_BUSINIT"; break;
        case 10: working = false; status = "NO_COMM_INV"; break;
        case 11: working = false; status = "SN_OVERCURRENT"; break;
        case 12: working = false; status = "BOOTLOAD"; break;
        case 13: working = false; status = "AFCI"; break;
        default: working = false; status = "UNKNOWN"; break;
        }
    if (istatus == 7) {
        x.error = tr("Status Fault, error code: %1: %2").arg(errcode).arg(stateToString(errcode));
        m_number_of_ok--;
        }

    x.invertor                   = invertor.invertor;
    x.now_power                  = (working) ? float32(rec, 51) : 0;
    x.now_ac_current             = (working) ? float32(rec, 11) : 0;
    x.now_ac_voltage             = 0; //??
    x.now_ac_frequency           = (working) ? float32(rec, 55) : 0;
    x.now_dc_current             = (working) ? float32(rec, 75) : 0;
    x.now_dc_voltage             = (working) ? float32(rec, 79) : 0;
    x.day_energy                 = 0;
    x.day_power_maximum          = 0;
    x.day_ac_voltage_maximum     = 0;
    x.day_ac_voltage_minimum     = 0;
    x.day_dc_voltage_maximum     = 0;
    x.day_operating_hours        = 0;
    x.total_energy               = (working) ? float32(rec, 71) : 0;
    x.total_power_maximum        = 0;
    x.total_ac_voltage_maximum   = 0;
    x.total_ac_voltage_minimum   = 0;
    x.total_dc_voltage_maximum   = 0;
    x.total_operating_hours      = 0;
    x.status                     = status;

    emit data(x);
    m_number_of_ok++;

/*
    qDebug() << "ac total current    " << float32(rec, 11);
    qDebug() << "ac phase-a current  " << float32(rec, 15);
    qDebug() << "ac phase-b current  " << float32(rec, 19);
    qDebug() << "ac phase-c current  " << float32(rec, 23);
    qDebug() << "ac voltage phase AB " << float32(rec, 27);
    qDebug() << "ac voltage phase BC " << float32(rec, 31);
    qDebug() << "ac voltage phase CA " << float32(rec, 35);
    qDebug() << "ac voltage phase A0 " << float32(rec, 39);
    qDebug() << "ac voltage phase B0 " << float32(rec, 43);
    qDebug() << "ac voltage phase C0 " << float32(rec, 47);
    qDebug() << "ac power            " << float32(rec, 51);
    qDebug() << "ac frequency        " << float32(rec, 55);
    qDebug() << "aparent power       " << float32(rec, 59);
    qDebug() << "reactive power      " << float32(rec, 63);
    qDebug() << "power factor        " << float32(rec, 67);
    qDebug() << "production wH       " << float32(rec, 71);
    qDebug() << "DC current value    " << float32(rec, 75);
    qDebug() << "DC voltage value    " << float32(rec, 79);
    qDebug() << "DC power value      " << float32(rec, 83);
    qDebug() << "cabinet temperature " << float32(rec, 87);
    qDebug() << "coolant temperature " << float32(rec, 91);
    qDebug() << "trafo temperature   " << float32(rec, 95);
    qDebug() << "other temperature   " << float32(rec, 99);
    qDebug() << "operating state     " << integer16(rec, 103);
    qDebug() << "vendor operating st." << integer16(rec, 105);
    qDebug() << "event flags 0-31    " << integer32(rec, 107);
    qDebug() << "event flags 32-63   " << integer32(rec, 109);
    qDebug() << "event flags 32-63   " << integer32(rec, 113);
    qDebug() << "event flags 32-63   " << integer32(rec, 117);
    qDebug() << "event flags 32-63   " << integer32(rec, 121);
    qDebug() << "event flags 32-63   " << integer32(rec, 125);
*/

}


float MODBUSTCP::float32(const QByteArray& data, int offset) const {
    int ival0 = data[offset++];
    int ival1 = data[offset++];
    int ival2 = data[offset++];
    int ival3 = data[offset++];
    int ival = (ival0 << 24) + (ival1 << 16) + (ival2 << 8) + ival3;
    float value = *(float*)&ival;
    return value;
}


qint16 MODBUSTCP::integer16(const QByteArray& data, int offset) const {
    int ival0 = data[offset++];
    int ival1 = data[offset++];
    int ival = (ival0 << 8) + ival1;
    return ival;
}


qint32 MODBUSTCP::integer32(const QByteArray& data, int offset) const {
    int ival0 = data[offset++];
    int ival1 = data[offset++];
    int ival2 = data[offset++];
    int ival3 = data[offset++];
    int ival = (ival0 << 24) + (ival1 << 16) + (ival2 << 8) + ival3;
    return ival;
}


QString MODBUSTCP::stateToString(int state) {
    switch (state) {
        case 102:   return "AC voltage too high";
        case 103:   return "AC voltage too low";
        case 105:   return "AC frequenci too high";
        case 106:   return "AC frequenci too low";
        case 107:   return "AC grid outside the permissible limits";
        case 108:   return "Stand alone operation detected";
        case 112:   return "RCMU error";

        case 301:   return "Overcurrent (AC)";
        case 302:   return "Overcurrent (DC)";
        case 303:   return "Power stage set over temperature";
        case 304:   return "Internal temperature too high";
        case 306:   return "Intermediate circuit voltage too low for feeding energy into the grid";
        case 307:   return "DC input voltage too low for feeding energy into the grid";
        case 308:   return "Intermediate circuit overvoltage";
        case 309:   return "DC input voltage MPPT 1 too high";
        case 313:   return "DC input voltage MPPT 2 too high";

        case 401:   return "No communication with power stage set possible  ";
        case 406:   return "Power stage set temperature sensor faulty";
        case 407:   return "Internal temperature sensor faulty";
        case 408:   return "DC feeding into the grid detected";
        case 412:   return "Fixed voltage mode has been selected instead of MPP voltage mode and the fixed voltage has been set to too low or too high a value";
        case 415:   return "Safety cut out via option card or RECERBO has triggered ";
        case 416:   return "No communication possible between power stage set and control system";
        case 417:   return "Hardware ID problem";
        case 419:   return "Unique ID conflict";
        case 421:   return "HID range error";
        case 425:   return "No communication possible with the power stage set";
        case 426:   return "Possible hardware fault";
        case 427:   return "Possible hardware fault";
        case 428:   return "Possible hardware fault";
        case 431:   return "Software problem";
        case 436:   return "Functional incompatibility (one or more PC boards in the inverter are not compatible with each other, e.g. after a PC board has been replaced.";
        case 437:   return "Power stage set problem";
        case 438:   return "Functional incompatibility (one or more PC boards in the inverter in the inverter are not compatible with each other, e.g. after a PC board has been replaced)";
        case 443:   return "Intermediate circuit voltage too low or asymmetric";
        case 445:   return "Invalid limit value settings";
        case 447:   return "Insulation fault";
        case 448:   return "Neutral conductor not connected";
        case 450:   return "Guard cannot be found";
        case 452:   return "Memory error detected";
        case 453:   return "Short term grid voltage error";
        case 454:   return "Short term grid frequency error";
        case 456:   return "Anti-islanding function is no longer implemented correctly";
        case 457:   return "Grid relay sticking";
        case 459:   return "Error when recording the measuring signal for the insulation test";
        case 460:   return "Reference voltage source for the digital signal processor (DSP) is working out of tolerance";
        case 461:   return "Fault in the DSP data memory";
        case 462:   return "Error with DC feed monitoring routine";
        case 463:   return "Reversed AC polarity, AC connector inserted incorrectly";
        case 474:   return "RCMU sensor faulty";
        case 475:   return "Solar panel ground fault, insulation fault (connection between solar panel and ground)";
        case 476:   return "Driver supply voltage too low";
        case 480:   return "Functional incompatibility (one or more PC boards in the inverter are not compatible with each other, e.g. after a PC board has been replaced)";
        case 481:   return "Functional incompatibility (one or more PC boards in the inverter are not compatible with each other, e.g. after a PC board has been replaced)";
        case 482:   return "Startup incomplete";
        case 483:   return "Voltage UDC fixed on MPP2 string out of limits";
        case 485:   return "CAN transmit buffer is full";

        case 501:   return "Insulation error on the solar panels";
        case 509:   return "No energy fed into the grid in the past 24 hours";
        case 515:   return "No communication with filter possible";
        case 516:   return "No communication possible with the storage unit";
        case 517:   return "Derating caused by too high a temperature";
        case 558:   return "Functional incompatibility (one or more PC boards in the inverter are not compatible with each other, e.g. after a PC board has been replaced)";
        case 560:   return "Derating caused by over-frequency";
        case 566:   return "Arc detector switched off (e.g. during external arc monitoring)";

        case 705:   return "Conflict when setting the inverter number (e.g. number already assigned)";
        case 721:   return "EEPROM has been reinitialised or EEPROM is faulty";
        case 731:   return "Initialisation error – USB stick is not supported";
        case 732:   return "Over current on USB stick";
        case 733:   return "No USB stick connected";
        case 734:   return "Update file not recognised or not present";
        case 735:   return "Update file does not match the device, update file too old";
        case 736:   return "Write or read error occurred";
        case 738:   return "Log file cannot be saved (e.g. USB stick is write protected or full)";
        case 743:   return "Error occurred during update process";
        case 745:   return "Update file corrupt";
        case 751:   return "Time lost";
        case 752:   return "Real Time Clock module communication error";
        case 757:   return "Hardware error in the Real Time Clock module";
        case 758:   return "Internal error: Real Time Clock module is in emergency mode";
        case 766:   return "Emergency power de-rating has been activated";
        };

    return "Uknown status code";
}


QString MODBUSTCP::exceptionCodeToString(int exceptionCode) {
    switch(exceptionCode) {
        case 0x01: return "Illegal function";
        case 0x02: return "Illegal data address";
        case 0x03: return "Illegal data value";
        case 0x04: return "Slave device failure";
        case 0x05: return "Acknowledge";
        case 0x06: return "Slave device busy";
        case 0x07: return "Negative acknowledge";
        case 0x08: return "Memory parity error";
        case 0x0a: return "Gateway path unavailable";
        case 0x0b: return "Gateway target device failed to respond";
        }
    return "";
}


