/**
 * @file
 *  @author Petr Bravenec
 *  @copyright GNU Public License
 */

#include "invertors.h"
#include "fronius.h"
#include "static.h"


INVERTORS::INVERTORS(QCoreApplication *p) : QObject(p) {
    m_invertors.clear();
    m_timer = new QTimer(this);
    m_timer->setSingleShot(true);
    m_timer->setInterval(7000);
    connect(m_timer,    SIGNAL(timeout()),
            this,         SLOT(slotQuitInvertors()));
    connect(Static::db, SIGNAL(databaseChanged()),
            this,         SLOT(databaseChanged()));

    m_timerWorking = new QTimer(this);
    m_timerWorking->setSingleShot(true);
    m_timerWorking->setInterval(1000);
    connect(m_timerWorking,  SIGNAL(timeout()),
            this,              SLOT(slotCountWorking()));
    m_timerWorking->start();
}


/**
 * @brief Registers changes made in database
 *
 * Response to changes in database is delayed since database
 * tends to make changes in groups.
 */
void INVERTORS::databaseChanged() {
    m_timer->start();
}


/**
 * @brief Sends signal to invertors and wait a moment to close them
 */
void INVERTORS::slotQuitInvertors() {
    emit quitInvertors();
    QTimer::singleShot(5000, this, SLOT(closeLines()));
}


void INVERTORS::closeLines() {
    m_timerWorking->stop();
    m_timerWorking->setInterval(1000);
    QCoreApplication::processEvents();
    for (int i=0; i<m_invertors.size(); i++) {
        m_invertors.at(i)->wait(250);
        m_invertors.at(i)->terminate();
        delete m_invertors.at(i);
        }
    m_invertors.clear();
    openLines();
    m_timerWorking->start();
    qWarning() << "INVERTORS::closeLines(): configuration reloaded";
}


QList<INVERTOR_status> INVERTORS::status() {
    QList<INVERTOR_status> list;
    for (int i=0; i<m_invertors.size(); i++) {
        INVERTOR_status s = m_invertors.at(i)->status();
        for (int w=0; w<m_working.size(); w++) {
            if (m_working.at(w).line == s.line) {
                s.working   = m_working.at(w).working;
                s.invertors = m_working.at(w).invertors;
                }
            }
        list << s;
        }
    return list;
}


void INVERTORS::slotCountWorking() {
    m_working = Static::db->linesStatus();
    m_timerWorking->setInterval(30000);
    m_timerWorking->start();
}


/**
 * @brief Opens serial lines administered in database
 */
void INVERTORS::openLines() {
    QList<DBT_LINES> lines = Static::db->lines();
    for (int i=0; i<lines.size(); i++) {
        QList<DBT_INVERTORS> invertors = Static::db->invertors(lines.at(i).line);
        INVERTOR *f = INVERTOR::create(lines[i].type); // new FRONIUS();
        f->setLine(lines.at(i));
        f->setInvertors(invertors);
        f->start();
        connect (this,     SIGNAL(quitInvertors()),
                 f,          SLOT(quit()),
                 Qt::QueuedConnection);
        connect (f,     SIGNAL(data(DBT_DATA)),
                 this,  SLOT(slotStoreData(DBT_DATA)),
                 Qt::QueuedConnection);
        connect (this,          SIGNAL(storeData(DBT_DATA)),
                 Static::db,    SLOT(insertData(DBT_DATA)),
                 Qt::QueuedConnection);
        connect (f,        SIGNAL(loopFinished(int,int)),
                 Static::db, SLOT(commit()),
                 Qt::QueuedConnection);
        connect (f,        SIGNAL(    loopFinished(int,int)),
                 this,       SLOT(slotLoopFinished(int,int)),
                 Qt::QueuedConnection);
        connect (f,        SIGNAL(data(DBT_DATA)),
                 this,       SLOT(data(DBT_DATA)),
                 Qt::QueuedConnection);
        m_invertors << f;
        }
}


void INVERTORS::data(DBT_DATA x) {
    qDebug() << QString("%1 sun=%2 pwr=%3 err=%4")
            .arg(x.invertor)
            .arg(Static::slunce->alt())
            .arg(x.now_power.toInt())
            .arg(x.error)
            ;
    return;
    if (x.error.isEmpty()) {
        qDebug() << "invertor="      << x.invertor
                 << "power="         << x.now_power
                 << "dc current="    << x.now_dc_current
                 << "dc voltage="    << x.now_dc_voltage 
                 << "ac current="    << x.now_ac_current
                 << "ac voltage="    << x.now_ac_voltage
                 << "ac frequency="  << x.now_ac_frequency
                 << "\n" 
                 << "ac voltage max="    << x.day_ac_voltage_maximum 
                 << "ac voltage min="    << x.day_ac_voltage_minimum 
                 << "dc voltage max="    << x.day_dc_voltage_maximum 
                 << "powr max="          << x.day_power_maximum
                 << "hours="             << x.day_operating_hours
                 << "energy="            << x.day_energy
                 << "\n" 
                 << "ac voltage max="    << x.total_ac_voltage_maximum 
                 << "ac voltage min="    << x.total_ac_voltage_minimum 
                 << "dc voltage max="    << x.total_dc_voltage_maximum 
                 << "powr max="          << x.total_power_maximum
                 << "hours="             << x.total_operating_hours
                 << "energy="            << x.total_energy
            ;
      } else {
        qDebug() << "invertor=" << x.invertor
                 << "err="      << x.error
            ;
        }
}

void INVERTORS::slotLoopFinished(int number_of_ok, int number_of_err) {
    if (number_of_ok > 0) return;
    if (number_of_ok <= 0 || number_of_err <= 0) return;
    if (Static::slunce == NULL || Static::slunce->alt() < 0.0) return;

    Static::interfacebox->restart();
}

void INVERTORS::slotStoreData(DBT_DATA d) {
    if (Static::slunce == NULL || Static::slunce->alt() > -18.0) { // astronomický soumrak
        emit storeData(d);
        }
}

