/**
 * @file
 * @author Petr Bravenec
 * @copyright GNU Public License
 */

#ifndef __INVERTORS_H__
#define __INVERTORS_H__

#include <QList>
#include <QTimer>
#include <QObject>
#include <QCoreApplication>
#include "invertor.h"
#include "invertor_status.h"


/**
 * @brief Manages list of invertors and their threads
 *
 * Containter for list of active lines. Every line is stored in
 * it's own thread - virtual class INVERTOR.
 *
 * Connects data sent from invertor's thread using signal INVERTORS::data(DBT_DATA) to
 * database's slot insertData(DBT_DATA). Receives signal 
 * DATABASE::sigDatabaseChanged() in slot INVERTORS::databaseChanged() and after a 
 * short delay closes all open lines (class INVERTOR) and restart new instances based
 * on new configuration stored in database.
 *
 * Stores status of working lines in list of DBT_LINES_STATUS structures. The status
 * is updated every 30 seconds independently of invertors functionality. This allows
 * to update status in situations when no invertor connected to line worked and the
 * INVERTOR class thread was blocked and not responding.
 *
 */
class INVERTORS : public QObject {
    Q_OBJECT
  public:
    INVERTORS(QCoreApplication *);

    /**
     * @brief Returns list of current invertors status
     */
    QList<INVERTOR_status>  status();

  signals:
    /**
     * @brief Signal for invertors to quit
     * 
     * When the signal is emitted then:
     *  - invertor's slot quit() was called
     *  - the object waited a few seconds and then destroy all invertors.
     *  - call openLines() to read new configuration and start new 
     *    invertor's instances
     */
    void quitInvertors();

    /**
     * @brief Signal for database to store invertor data 
     */
    void storeData(DBT_DATA);

  public slots:
    /**
     * @brief Opens all configured invertors
     *
     * Every invertor is opened in it's own thread
     * so more invertors can work at the same time.
     */
    void openLines();

    /**
     * @brief Closes all existing invertors and destoroy their instances
     */
    void closeLines();

    /**
     * @brief Slot for handling of requests for storing of invertor data in database
     */
    void slotStoreData(DBT_DATA);

  private slots:
    /**
     * @brief Slot is called when configuration changed in database
     */
    void databaseChanged();

    /**
     * @brief Count working invertors
     */
    void slotCountWorking();

    /**
     * @brief Slot is called a few seconds after configuration change.
     *
     * It sends signal quitInvertors() and after a short pause (timer)
     * calls closeLines() to destroy invertors.
     *
     */
    void slotQuitInvertors();

    /**
      * @brief Slot is called when the line finished the reading cycle
      *
      * Slot receives number of successfull data and error data.
      * When all the invertors are in error, then the interfacebox is restarted
      *
      */
    void slotLoopFinished(int number_of_ok, int number_of_err);

    /**
     * @brief For debugging only, writes read data to console
     */
    void data(DBT_DATA);

  private:
    QList<INVERTOR*>    m_invertors;    ///< List of all invertors
    QTimer             *m_timer;        ///< Timer for database changes responses

    QTimer             *m_timerWorking; ///< Timer for periodic status update
    QList<DBT_LINES_STATUS> m_working;  ///< List of working invertors
};

#endif
