#include "slunce.h"
#include <math.h>
#include <QMutexLocker>
#include <QTimer>

SLUNCE::~SLUNCE() {
}


SLUNCE::SLUNCE(QCoreApplication *) : QThread(0) {
    obs_long = 18;
    obs_lat  = 49;
    datum    = QDateTime::currentDateTime();
    moveToThread(this);
}


void SLUNCE::start(Priority priority) {
    QTimer::singleShot(10, this, SLOT(slotInit()));
    QThread::start(priority);
}


void SLUNCE::slotInit() {
    spocti();
    QTimer *t = new QTimer(this);
    t->setSingleShot(false);
    t->setInterval(60000); // minuta
    t->start();
    connect(t, SIGNAL(timeout()), this, SLOT(spocti()) );
}


/**
 * Nastaví zeměpisnou polohu pozorovatele (panelu).
 *
 * @param lat - zeměpisná šířka ve stupních, sever kladně
 * @param lng - zeměpisná délka ve stupních, východ kladně
 */
void SLUNCE::setObserver(double lat, double lng) {
    obs_lat  = lat;
    obs_long = lng;
}


void SLUNCE::spocti() {
    QMutexLocker locker(&m_mutex);
    datum = QDateTime::currentDateTime();
    sun_ra_dec();
    sun_alt_az();
}


double SLUNCE::ra() {
    QMutexLocker locker(&m_mutex);
    return sun_ra / 15.0;
}


double SLUNCE::dec() {
    QMutexLocker locker(&m_mutex);
    return sun_dec; 
}


double SLUNCE::alt() {
    QMutexLocker locker(&m_mutex);
    return sun_alt; 
}


double SLUNCE::az() {
    QMutexLocker locker(&m_mutex);
    return sun_az ; 
}


void SLUNCE::sun_ra_dec() {

    // do proměnné d vypočítej počet dní (na desetinná místa) od 1.1.2000 12:00 gmt
    QDateTime epoch = QDateTime(QDate(2000,1,1),QTime(12,00),Qt::UTC);
    datum_jd = ((double)epoch.secsTo( datum )) / 86400 ;

    double mean_longitude     = to360 (280.461 + 0.9856474 * datum_jd);
    double mean_anomaly       = to360 (357.528 + 0.9856003 * datum_jd);
    double ecliptic_longitude = to360 (mean_longitude + 1.915 * sin(mean_anomaly/DTR) + 0.020 * sin(2*mean_anomaly/DTR));
    double ecliptic_obliquity = to360 (23.429 - 0.0000004 * datum_jd);
    double Y = cos(ecliptic_obliquity/DTR) * sin(ecliptic_longitude/DTR);
    double X = cos(ecliptic_longitude/DTR);

    // rektascenze je OK, pozor na různé jednotky (hodiny, stupně)
    sun_ra  = DTR * atan2 (Y, X);
    sun_dec = DTR * asin (sin(ecliptic_obliquity/DTR) * sin(ecliptic_longitude/DTR));

}


void SLUNCE::sun_alt_az() {
    double gmt_mean_sidereal_time ;
    double local_mean_sidereal_time ;
    double hour_angle;

    double date = floor(datum_jd);
    double hour = (datum_jd - date - 0.5) * 24;

    gmt_mean_sidereal_time   = to24 ( 6.697375 + 0.0657098242 * date + hour);
    local_mean_sidereal_time = to24 ( gmt_mean_sidereal_time + obs_long / 15. );
    hour_angle               = to360( local_mean_sidereal_time * 15.0 - sun_ra );

    sun_alt = sin(sun_dec/DTR) * sin(obs_lat/DTR) + cos(sun_dec/DTR) * cos(obs_lat/DTR) * cos(hour_angle/DTR);
    sun_alt = asin(sun_alt) * DTR;

    sun_az  = (sin(sun_dec/DTR) - sin(sun_alt/DTR) * sin(obs_lat/DTR) ) /
              (cos(sun_alt/DTR) * cos(obs_lat/DTR) );
    sun_az  = acos(sun_az) * DTR;
    if (sin(hour_angle/DTR)>0) { sun_az = 360 - sun_az; }

}

QString SLUNCE::toDeg(double x) {
    double d  = floor(x);
    double mf = (x - d) * 60;
    double m  = floor(mf);
    double s  = (mf - m) * 60;
    return QString(QString("%1:%2:%3").arg(d).arg(m).arg(s));
}

double SLUNCE::to360(double x) {
    while (x < 0)   { x = x + 360; }
    while (x > 360) { x = x - 360; }
    return x;
}

double SLUNCE::to24(double x) {
    while (x < 0)   { x = x + 24; }
    while (x > 24)  { x = x - 24; }
    return x;
}

