/**
 * @file
 * @author Petr Bravenec
 * @copyright GNU Public License
 */

#include "static.h"
#include "startup.h"
#include "dualfilelogger.h"
#include "httplistener.h"
#include "requestmapper.h"
#include "staticfilecontroller.h"
#include "invertors.h"
#include "maintenance.h"
#include <QLibraryInfo>
#include <QTranslator>
#include <QTextCodec>
#include <QtGlobal>

/** Name of this application */
#define APPNAME "fotobot"

/** Publisher of this application */
#define ORGANISATION "Hobrasoft"

/** Short description of this application */
#define DESCRIPTION "Applications gets data from fronius invertors"


void Startup::start() {
    // Initialize the core application
    QCoreApplication* app = application();
    app->setApplicationName(APPNAME);
    app->setOrganizationName(ORGANISATION);
    QString configFileName=Static::getConfigFileName();

    // Configure logging into files
    QSettings* mainLogSettings=new QSettings(configFileName,QSettings::IniFormat,app);
    mainLogSettings->beginGroup("mainLogFile");
    QSettings* debugLogSettings=new QSettings(configFileName,QSettings::IniFormat,app);
    debugLogSettings->beginGroup("debugLogFile");
    Logger* logger=new DualFileLogger(mainLogSettings,debugLogSettings,10000,app);
    Q_UNUSED(logger);
//  logger->installMsgHandler();

    // Translations
    QSettings localeSettings(configFileName, QSettings::IniFormat, app);
    localeSettings.beginGroup("locale");
    QString locale = localeSettings.value("locale", QLocale::system().name()).toString();
    QTextCodec::setCodecForTr(QTextCodec::codecForName("utf8"));
    QTranslator *translator = new QTranslator;
    QString an = app->applicationName();
    QString lp = QLibraryInfo::location(QLibraryInfo::TranslationsPath);
    QString tf = an  + "_" + locale;
    QString l1 = "/usr/share/locale/";
    QString l2 = "/usr/local/share/locale/";
    QString l3 = QString("/usr/share/") + an;
    QString tx = locale + "/LC_MESSAGES/" + an;
    for (;;) {
        if (translator->load (tf, lp ) ) break;
        if (translator->load (tf, ".") ) break;
        if (translator->load (tf, an ) ) break;
        if (translator->load (tx, l1 ) ) break;
        if (translator->load (tx, l2 ) ) break;
        if (translator->load (tf, l3 ) ) break;
        qWarning() << "Translations not loaded";
        break;
        }
    app->installTranslator(translator);

    // Configure database file
    QSettings* databaseSettings = new QSettings(configFileName,QSettings::IniFormat,app);
    databaseSettings->beginGroup("database");
    Static::db = new DATABASE(databaseSettings, app);
    new MAINTENANCE();

    // Configure invertors
    Static::invertors = new INVERTORS(app);
    Static::invertors->openLines();

    // Startup interacebos restarter
    Static::interfacebox = new INTERFACEBOX(app);
    Static::interfacebox->setDatabase(Static::db);
    Static::interfacebox->start();

    // Startup sun position calculations
    Static::slunce = new SLUNCE(app);
    DBT_GEOCOORDINATES geo;
    geo = Static::db->geocoordinates();
    Static::slunce->setObserver(geo.latitude.toDouble(), geo.longitude.toDouble());
    Static::slunce->start();

    // Configure template loader and cache
    QSettings* templateSettings=new QSettings(configFileName,QSettings::IniFormat,app);
    templateSettings->beginGroup("templates");
    Static::templateLoader=new TemplateCache(templateSettings,app);

    // Configure session store
    QSettings* sessionSettings=new QSettings(configFileName,QSettings::IniFormat,app);
    sessionSettings->beginGroup("sessions");
    Static::sessionStore=new HttpSessionStore(sessionSettings,app);

    // Configure static file controller
    QSettings* fileSettings=new QSettings(configFileName,QSettings::IniFormat,app);
    fileSettings->beginGroup("docroot");
    Static::staticFileController=new StaticFileController(fileSettings,app);

    // Configure and start the TCP listener
    QSettings* listenerSettings=new QSettings(configFileName,QSettings::IniFormat,app);
    listenerSettings->beginGroup("listener");
    new HttpListener (listenerSettings, new RequestMapper(app), app);

    qWarning("Service has started");
}


void Startup::stop() {
    qWarning("Service has been stopped");
    // QCoreApplication destroys all objects that have been created in start().
}


Startup::Startup(int argc, char *argv[])
    : QtService<QCoreApplication>(argc, argv, APPNAME) {
    setServiceDescription(DESCRIPTION);
    //setStartupType(AutoStartup);
}



