/**
 * @file
 * @copyright GNU Public License
 * @author Petr Bravenec, Jiří Procházka
*/

#include "linescontroller.h"
#include "invertorscontroller.h"
#include "static.h"
#include <QVariant>
#include <QDateTime>

LinesController::LinesController() {
    qRegisterMetaType<DBT_LINES>();
    connect(this, SIGNAL(saveLine(DBT_LINES)), Static::db, SLOT(saveLine(DBT_LINES)), Qt::BlockingQueuedConnection);
    connect(this, SIGNAL(deleteLine(int)), Static::db, SLOT(deleteLine(int)), Qt::QueuedConnection);
    qRegisterMetaType<DBT_INVERTORS>();
    connect(this, SIGNAL(saveInvertor(DBT_INVERTORS)), Static::db, SLOT(saveInvertor(DBT_INVERTORS)), Qt::QueuedConnection);
}

void LinesController::servicePrivate(HttpRequest& request) {
    QByteArray action = request.getParameter("action");
    if(action.startsWith("add")) {
        actionAdd(request); return;
    } else if(action.startsWith("edit")) {
        actionEdit(request); return;
    } else if(action.startsWith("delete")) {
        actionDelete(request); return;
    }

    /** Default action/page */

    QList<DBT_LINES> lines = Static::db->lines();
    write(QString("<h2>%1</h2>\n").arg(tr("Lines")));
    write(QString("<a href=\"/lines?action=add\">%1</a>\n").arg(tr("Add a line")));
    write("<table>\n");
    write(QString("<tr><th></th><th>%1</th><th>%2</th><th>%3</th><th>%4</th><th>%5</th><th>%6</th><th>%7</th><th>%8</th></tr>\n").arg(tr("Line"), tr("Device"), tr("IP address"), tr("Description"), tr("Serial Number Format"), tr("Speed"), tr("Retries"), tr("Timeout [ms]")));

    QList<DBT_LINES>::const_iterator i;
    for(i = lines.constBegin(); i != lines.constEnd(); ++i) {
        const DBT_LINES& l = *i;
        write(QString("<tr><td>"
                "<a href=\"/lines?action=edit&id=%1\">%9</a> "
                "<a href=\"/lines?action=delete&id=%1\">%10</a>"
            "</td>"
            "<td>%1</td><td>%2</td><td>%3</td><td>%4</td><td>%5</td><td>%6</td><td>%7</td><td>%8</td>"
            "</tr>\n")
            .arg(l.line)        // 1
            .arg(l.device)      // 2
            .arg(l.hostname + ":" + QString::number(l.portnumber) )
            .arg(l.description )  // 4
            .arg(l.sn_format)   // 5
            .arg(l.speed)       // 6
            .arg(l.retries)     // 7
            .arg(l.timeout)     // 8
            .arg(tr("Edit"))    // 9
            .arg(tr("Delete"))  // 10
            );
    }

    write("</table>\n");
}

QString LinesController::getEditFormContents(const DBT_LINES& line) {
    QString s;
    s.append(QString("<input type=\"hidden\" name=\"id\" value=\"%1\" />\n").arg(line.line));
    QString i ("<tr><td><label for=\"%2\">%1</label>: </td><td><input type=\"text\" name=\"%2\" value=\"%3\" /></td></tr>\n");
    s.append(i.arg(tr("Description"), "description", line.description));
    s.append(i.arg(tr("Serial Number Format"), "sn_format", line.sn_format));

    QString sel;
    sel += QString("<option value=\"%1\" %2>%3</option>")
                .arg(DBT_LINES::Fronius)
                .arg((line.type == DBT_LINES::Fronius)?"selected=\"selected\"" : "")
                .arg("Fronius (interfacebox)")
                ;
    sel += QString("<option value=\"%1\" %2>%3</option>")
                .arg(DBT_LINES::ModbusTCP)
                .arg((line.type == DBT_LINES::ModbusTCP)?"selected=\"selected\"" : "")
                .arg("Modbus TCP")
                ;
    s.append(QString("<tr><td>%1</td><td><select id=\"type\" name=\"type\">%2</select></td>")
                .arg(tr("Type"))
                .arg(sel)
                );

    s.append(i.arg(tr("Device"), "device", line.device));
    s.append(i.arg(tr("Hostname"), "hostname", line.hostname));
    s.append(i.arg(tr("Port number"), "portnumber", QString::number(line.portnumber)));
    // speed selectbox
    s.append(QString("<tr><td><label for=\"%2\">%1</label>: </td><td><select id=\"%2\" name=\"%2\">\n").arg(tr("Speed"), "speed"));
    QList<int> speeds;
    speeds << 110 << 300 << 600 << 1200 << 2400 << 4800 << 9600 << 19200 << 38400 << 57600 << 115200;
    QList<int>::const_iterator ii;
    for(ii = speeds.constBegin(); ii != speeds.constEnd(); ++ii) {
        int sp = *ii;
        s.append(QString("<option value=\"%1\" %2>%1</option>\n")
                .arg(sp).arg((sp == line.speed) ? "selected=\"selected\"" : ""));
    }
    s.append("</select></td></tr>\n");

    s.append(i.arg(tr("Retries"), "retries").arg(line.retries));
    s.append(i.arg(tr("Timeout [ms]"), "timeout").arg(line.timeout));
    return s;
}

void LinesController::readForm(DBT_LINES& line, HttpRequest& request) {
    line.line = request.getParameter("id").toInt();
    line.device = request.getParameter("device");
    line.description = request.getParameter("description");
    line.sn_format = request.getParameter("sn_format");
    QByteArray p = request.getParameter("speed");
    line.speed = (p.isEmpty()) ? 9600 : p.toInt();
    p = request.getParameter("retries");
    line.retries = (p.isEmpty()) ? 5 : p.toInt();
    p = request.getParameter("timeout");
    line.timeout = (p.isEmpty()) ? 3000 : p.toInt();
    p = request.getParameter("type");
    line.type    = (DBT_LINES::Type)(
            (p.isEmpty()) ? DBT_LINES::Fronius : p.toInt()
            );
    line.hostname   = request.getParameter("hostname");
    line.portnumber = request.getParameter("portnumber").toInt();
}

/**
 * Add a line action
 */
void LinesController::actionAdd(HttpRequest& request) {
    QByteArray action = request.getParameter("action");
    /// a form
    write(QString("<h3>%1</h3>\n").arg(tr("Add a line")));
    write("<form method=\"post\">\n"
        "<input type=\"hidden\" name=\"action\" value=\"add2\" />\n");
    DBT_LINES line;
    readForm(line, request);
    write("<table class=\"formTable\">");
    write(getEditFormContents(line));

    /// invertor generator form
    write(QString("<tr><td><h4>%1</h4></td><td></td></tr>").arg(tr("Generate Invertors")));
    DBT_INVERTORS invertor;
    InvertorsController::readForm(invertor, request);
    if(invertor.address < 1) invertor.address = 1;
    write(InvertorsController::getEditFormContents(request, invertor, true));
    write("</table>");

    write(QString("<input type=\"submit\" value=\"%1\">\n"
    "</form>\n").arg(tr("Add")));
    if(action == "add2") {
        /// second page doing the action and redirecting
        line.line = 0;
        invertor.invertor = 0;
        /// validation
        if(true) {
            invertor.line = Static::db->saveLine(line);

            /// invertor generation
            QByteArray generate = request.getParameter("generate");
            int gen = (generate.isEmpty()) ? 0 : generate.toInt();
            for(int j = 0; j < gen; j++) {
                emit saveInvertor(invertor);
                ++invertor.address; // do addresses have to be unique only in 1 line?
            }

            QString location = QString("/lines?message=%1").arg(percentEncode(tr("Line added.")));
            redirect(location);
        } else {
            write(QString("<p>%1</p>").arg(tr("Line definition invalid.")));
        }
    }
}

/**
 * Edit line action
 */
void LinesController::actionEdit(HttpRequest& request) {
    QByteArray action = request.getParameter("action");
    DBT_LINES line;
    /// a form
    write(QString("<h3>%1</h3>\n").arg(tr("Edit a line")));
    write("<form method=\"post\">\n"
        "<input type=\"hidden\" name=\"action\" value=\"edit2\" />\n");
    if(action == "edit") {
        QByteArray id = request.getParameter("id");
        int idi = id.toInt();
        if(id.isEmpty() || idi < 1) {
            redirect(QString("/lines?message=%1").arg(percentEncode(tr("Error: Invalid line ID."))));
            return;
        }
        QList<DBT_LINES> lines = Static::db->lines(idi);
        if(lines.empty()) {
            redirect(QString("/lines?message=%1").arg(percentEncode(tr("Error: no such line found."))));
            return;
        }
        write("<table class=\"formTable\">");
        write(getEditFormContents(lines.at(0)));
        write("</table>");
    } else {
        readForm(line, request);
        write("<table class=\"formTable\">");
        write(getEditFormContents(line));
        write("</table>");
    }
    write(QString("<input type=\"submit\" value=\"%1\">\n"
    "</form>\n").arg(tr("Edit")));
    if(action == "edit2") {
        /// second page doing the action and redirecting
        /// validation
        if(line.line > 0 && true) {
            emit saveLine(line);
            QString location = QString("/lines?message=%1").arg(percentEncode(tr("Line edited.")));
            redirect(location);
        } else {
            write(QString("<p>%1</p>").arg(tr("Error: Line definition invalid.")));
        }
    }
}

/**
 * Delete line action
 */
void LinesController::actionDelete(HttpRequest& request) {
    write(QString("<h3>%1</h3>\n").arg(tr("Delete a line")));
        QString location;
        QByteArray id = request.getParameter("id");
        int idi = id.toInt();
        if(!id.isEmpty() && idi > 0) {
            emit deleteLine(idi);
            location = QString("/lines?message=%1").arg(percentEncode(tr("Line deleted.")));
        } else {
            location = QString("/lines?message=%1").arg(percentEncode(tr("Error: Invalid line ID. Line not deleted.")));
        }
        redirect(location);
}
