/**
 * @file
 * @author Petr Bravenec, Jiří Procházka
 * @copyright GNU Public License
 */

#include "webpage.h"
#include "static.h"
#include "version.h"
#include <QVariant>
#include <QDateTime>

WebPage::WebPage() : MRequestHandler() {
    m_needs_access_data     = true;
    m_needs_change_settings = true;
    m_response    = NULL;
    m_cacheable   = false;
}

void WebPage::service(HttpRequest& request, HttpResponse& response) {
    m_response = &response;
    m_request  = &request;
    
    if(!m_cacheable) {
        response.setHeader("Cache-Control", "private, must-revalidate, max-age=0");
        response.setHeader("Pragma", "no-cache");
        }

    if (needsLogin(request,response)) {
        return;
        }

    servicePrivate  (request);
    writeHeader     (request,response);
    response.write  (m_body);
    writeFooter     (request,response);
}

void WebPage::redirect(const QString& location) {
    setStatus(301, "Moved Permanently");
    setHeader("Location", location.toUtf8());
    setHeader("Content-Type", "text/html; charset=utf-8");
    write(tr("<h1>Moved Permanently</h1><p>The document has moved <a href=\"%1\">here</a>.</p>\n").arg(location));
}

QString WebPage::percentEncode(const QString& s) {
    return QString(s.toUtf8().toPercentEncoding());
}

void WebPage::writeFooter(HttpRequest& request, HttpResponse& response) {
    Q_UNUSED(request);
    response.write("</div><!-- content -->\n"
        "</div><!-- page -->\n"
        "</body></html>");
}

/**
 * Aside from drawing login status, displays message from HTTP parameter "message" (percent-encoded)
 */
void WebPage::writeHeader(HttpRequest& request, HttpResponse& response) {
    QString body ("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\"\n"
        "   \"http://www.w3.org/TR/html4/strict.dtd\">\n"
        "<html><head>\n"
        "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">\n"
        "<title>Fotobot</title>\n"
        "<link href=\"/style.css\" rel=\"stylesheet\" type=\"text/css\">\n"
        "<script type=\"text/javascript\">\n");
    body.append(m_javascript);
    body.append(tr("</script>\n"
        "</head>\n"
        "<body>\n"
        "<div id=\"page\">\n"
        "<div id=\"header\">\n"
        "<div id=\"logo\">\n"
        "<h1>Fotobot</h1>\n"
        "<p>Collects data from your photovoltaic plant<br/>Made by <a href=\"http://hobrasoft.cz/\">Hobrasoft s.r.o.</a></p>\n"
        "<p class=\"clear\"></p></div><!-- logo -->\n"
        "<div id=\"menu\">\n"
        "<ul>\n"
        ));
    body.append(QString("<li><a href=\"/\">%1</a></li>\n").arg(tr("Main Page")));
    body.append(QString("<li><a href=\"/export\">%1</a></li>\n").arg(tr("Current data")));
    body.append(QString("<li><a href=\"/settings\">%1</a><ul>\n").arg(tr("Settings")));
    body.append(QString("<li><a href=\"/lines\">%1</a></li>\n").arg(tr("Lines")));
    body.append(QString("<li><a href=\"/invertors\">%1</a></li>\n").arg(tr("Invertors")));
    body.append(QString("<li><a href=\"/interfaceboxes\">%1</a></li>\n").arg(tr("Interfaceboxes")));
    body.append(QString("<li><a href=\"/interfaceboxes?action=restart\">%1</a></li>\n").arg(tr("Interfaceboxes restart")));
    body.append(QString("<li><a href=\"/geocoordinates\">%1</a></li>\n").arg(tr("Geographical coordinates")));
    #ifdef GENTOO
    body.append(QString("<li><a href=\"/network\">%1</a></li>\n").arg(tr("Network settings")));
    #endif
    body.append(QString("<li><a href=\"/login?action=chpw\">%1</a></li>\n").arg(tr("Change Password")));
    body.append(QString("<li><a href=\"/users\">%1</a></li>\n").arg(tr("Users")));
    body.append("</ul></li>\n");

    HttpSession session=Static::sessionStore->getSession(request,response);
    if (session.contains("username")) {
        body.append(QString("<li><a href=\"/login?logout=true\">%1</a></li>").arg(tr("Log out")));
    } else {
        body.append(QString("<li><a href=\"/login\">%1</a></li>\n").arg(tr("Log in")));
    }

    body.append("</ul><p class=\"clear\"></p></div><!-- menu -->\n");
    body.append("</div><!-- header -->\n");

    QString message = QString::fromUtf8(QByteArray::fromPercentEncoding(request.getParameter("message")));
    if(!message.isEmpty()) {
        body.append(QString("<p class=\"message\">%1</p>").arg(message));
    }

    body.append("<div id=\"content\">");
    response.write(body.toUtf8());
}


void WebPage::about() {
    write(QString("<h2>%1</h2>\n").arg(tr("About")));
    write(QString("<p>%1 %2</p>\n").arg(tr("Version")).arg(VERSION));
    write(tr(
        "<p>Application Fotobot is made by <a href=\"http://hobrasoft.cz\">Hobrasoft s.r.o.</a>\n"
        "It is a part of complex photovoltaic suite:</p>\n"
        "<ul>\n"
        "<li><a href=http://hobrasoft.cz/en/fotomon/fotomon>Fotomon</a> - monitoring, analyses and report making software for central monitoring services.</li>\n"
        "<li><a href=http://fm.solartec.cz/>Fotoweb</a> - web for photovoltaic system owner, a lot of graphs, businness information and detailed operating log for every photovoltaic plant.</li>\n"
        "<li><a href=http://hobrasoft.cz/en/fotomon/fotobot>Fotobot</a> - simple but very quick application to collect data from your photovoltaic plant.</li>\n"
        "<li><a href=http://hobrasoft.cz/en/fotobothw>FotobotHW</a> - small but powerfull computer to collect data from photovoltaic plant. Based on Linux, BeagleBone, Hobrasoft 4&nbsp;&times;&nbsp;RS232</li>\n"
        "</ul>\n"
        "<p>Home page: <a href=\"http://hobrasoft.cz/fotobot\">http://hobrasoft.cz/fotobot</a></p>\n"
        "<p>Application Fotobot is available under <a href=\"http://www.gnu.org/copyleft/gpl.html\">GPL license</a></p>"
        ));
}

